# this is where the stack class is implemented

from __future__ import print_function

import copy
import os
import stat
import shutil
import subprocess
import sys
import time
from collections import defaultdict

import yaml

import requests

from cloudmesh_client.common.Shell import Subprocess, SubprocessError
from cloudmesh_client.shell.console import Console
from cloudmesh_client.db import CloudmeshDatabase
from pprint import pprint
from cloudmesh_client.default import Default

requests.packages.urllib3.disable_warnings()


def get_virtualenv_environment(venvpath):
    """Figures out the environment variables that are set when activating
    a virtualenv

    Example:
    >>> os.system('virtualenv /tmp/venv')
    >>> get_virtualenv_environment('/tmp/venv')


    :param venvpath: path to the virtual environment
    :returns: dictionary of environment variables
    :rtype: :class:`dict`

    """

    command = 'source %s/bin/activate' % venvpath
    script_lines = [
        'source {venvpath}/bin/activate >/dev/null 2>&1',
        '{command} >/dev/null 2>&1',
        'env',
    ]
    script = ';'.join(script_lines).format(**locals())
    output = subprocess.check_output(['bash', '-c', script])

    env = dict()
    for line in output.split('\n'):
        if '=' not in line:
            continue
        k, v = line.strip().split('=', 1)
        env[k] = v

    return env


class ProjectDB(object):

    filename = '.cloudmesh_projectdb.yml'
    default_name = 'p-'

    def __init__(self, prefix='~/.cloudmesh/projects'):
        prefix = os.path.abspath(
            os.path.expanduser(
                os.path.expandvars(
                    prefix
                )
            )
        )

        if os.path.exists(os.path.join(prefix, self.filename)):
            yp = os.path.join(prefix, self.filename)
            with open(yp) as fd:
                y = yaml.load(fd)
            self.__dict__.update(y)
            self.path = prefix

        else:

            if os.path.isfile(prefix):
                raise OSError('`{}` is a file, should be a directory'
                              .format(prefix))

            if not os.path.isdir(prefix):
                os.makedirs(prefix)

            self.path = prefix
            self.active = None
            self.generated_pid = 0

    def __iter__(self):
        for name in os.listdir(self.path):
            projdir = os.path.join(self.path, name)
            projfile = os.path.join(projdir, Project.filename)
            if os.path.isdir(projdir) and os.path.isfile(projfile):
                project = Project.load(projdir)
                yield project

    def __getitem__(self, projname):
        return Project.load(os.path.join(self.path, projname))

    def lookup(self, projname):
        """Lookup the project with the given name.

        If `projname` is None, this returns the currently active project.

        :param projname: Project name
        :returns: the project
        :rtype: :class:`Project`
        """
        if projname is None:
            return self.getactive()
        else:
            return self[projname]

    def add(self, project, force=False, update=False):
        """Add a project to the database

        :param project: the project to add_from_path
        :type project: :class:`Project`
        :param force: whether or not to force initialiation
        :type force: :class:`bool`
        :param update: whether or not to update a previous initialied project
        :type update: :class:`bool`
        """
        projdir = self.projectdir(project.name)
        if os.path.exists(projdir) and not force:
            raise ValueError('Project {} already exists: {}'
                             .format(project.name, projdir))

        project.init(force=force, update=update)
        project.sync_metadata(projdir)

    def sync_metadata(self, projects=True):
        """Synchronize the metadata to the backend store

        :param bool projects: project metadata should be saved as well.
        :type projects: :class:`bool`
        """
        if projects:
            for project in self:
                self.update(project)

        prop = copy.copy(self.__dict__)
        prop.pop('path')
        y = yaml.dump(prop, default_flow_style=False)
        with open(os.path.join(self.path, ProjectDB.filename), 'w') as fd:
            fd.write(y)

    def update(self, project):
        """Update the database with the given project

        :param project: the project
        :type project: :class:`Project`
        """
        path = os.path.join(self.path, project.name)
        project.sync_metadata(path)

    def projectdir(self, name):
        """Get the project directory.

        :param name: the name of a project
        :type name: :class:`str`
        :returns: the path to the project
        :rtype: :class:`str`
        """
        return os.path.join(self.path, name)

    def new_project_name(self):
        """Automatically generate a new project name

        :returns: a project name
        :rtype: :class:`str`
        """

        pid = self.generated_pid
        self.generated_pid += 1

        name = '{}{}'.format(self.default_name, pid)
        assert not os.path.exists(os.path.join(self.path, name))
        self.sync_metadata(projects=False)
        return name

    def activate(self, project):
        """Activate the given project

        :param project: the project to make active
        :type project: :class:`Project`
        """
        self.active = project.name
        self.sync_metadata(projects=False)

    def isactive(self, project):
        """Predicate indicating activation status of the project

        :param project: the project to check
        :type project: :class:`Project`
        :rtype: :class:`bool`
        """
        return self.active == project.name

    def getactive(self):
        """Returns the currently active project.

        :returns: the currently active project
        :rtype: subclass of :class:`Project`
        :raises: :class:`KeyError` if no project is currently active
        """

        return self[self.active]


class ProjectFactory(object):
    """Outer API used to create projects.

    This factory should be used rather than directly constructing
    :class:`Project` instances.

    """

    def __init__(self, prefix='~/.cloudmesh/projects'):
        path = os.path.abspath(os.path.expanduser(os.path.expandvars(prefix)))
        self.db = ProjectDB(path)
        self.stacktype = 'bds'
        self.project_name = None
        self.repo = None
        self.branch = None
        self.overrides = None
        self.playbooks = None
        self.force = False
        self.udpate = False

    def __call__(self):
        name = self.project_name or self.db.new_project_name()
        projdir = self.db.projectdir(name)

        if self.stacktype == 'bds':
            kwargs = dict()
            if self.repo:
                kwargs['repo'] = self.repo
            if self.branch:
                kwargs['branch'] = self.branch
            stack = BigDataStack(projdir, **kwargs)
        else:
            raise NotImplementedError('Unknown stack type {}'
                                      .format(self.stacktype))

        deployparams = dict()

        if self.username:
            deployparams['user'] = self.username

        if self.ips:
            deployparams['ips'] = self.ips

        if self.overrides:
            deployparams['overrides'] = self.overrides

        if self.playbooks:
            deployparams['playbooks'] = self.playbooks

        project = Project(name, stack, deployparams)
        self.db.add(project, force=self.force, update=self.udpate)

        if self.activate:
            self.db.activate(project)

        return project

    def use_bds(self):
        """Use the BigDataStack backend"""
        Console.debug_msg('Factory use_bds')
        self.is_bds = True
        return self

    def set_user_name(self, username):
        """Set the cluster login user name"""
        Console.debug_msg('Factory set_user_name: {}'.format(username))
        self.username = username
        return self

    def set_project_name(self, name):
        """Set the project name"""
        Console.debug_msg('Factory set_project_name: {}'.format(name))
        self.project_name = name
        return self

    def set_ips(self, ips):
        """Set the cluster IP addresses"""
        Console.debug_msg('Factory set_ips: {}'.format(ips))
        assert len(ips) >= 0
        self.ips = ips
        return self

    def set_repo(self, repo):
        """Set the repository to get the stack from"""
        Console.debug_msg('Factory set_repo: {}'.format(repo))
        self.repo = repo
        return self

    def set_branch(self, branch='master'):
        """Set the branch of the repository to use"""
        Console.debug_msg('Factory set_branch: {}'.format(branch))
        self.branch = branch
        return self

    def set_overrides(self, overrides=None):
        """Set the overrides for deployment"""
        Console.debug_msg('Factory set_overrides: {}'.format(overrides))
        overrides = overrides or defaultdict(lambda: defaultdict(list))
        self.overrides = overrides
        return self

    def set_playbooks(self, playbooks=None):
        """Set the playbooks to deploy"""
        Console.debug_msg('Factory set_playbooks: {}'.format(playbooks))
        playbooks = playbooks or list()
        self.playbooks = playbooks
        return self

    def set_force(self, force=False):
        """Set the `force` parameter"""
        Console.debug_msg('Factory set_force: {}'.format(force))
        self.force = force
        return self

    def set_update(self, update=False):
        """Set the `update` parameter"""
        Console.debug_msg('Factory set_update: {}'.format(update))
        self.update = update
        return self

    def activate(self, make_active=True):
        """Set whether or not the created project should be activated"""
        Console.debug_msg('Factory activate: {}'.format(make_active))
        self.make_active = make_active
        return self


class Project(object):
    """Captures the parameters and stack for a deployment.  Note: this is
    a lower-level API that should be predominantly be constructed
    using the :class:`ProjectFactory`.
    """

    filename = '.cloudmesh_project.yml'

    def __init__(self, name, stack, deployparams=None):
        assert deployparams is not None

        self.name = name
        self.ctime = time.gmtime()
        self.stack = stack
        self.is_deployed = False
        self.deployparams = deployparams
        self.deployparams['name'] = name

    @classmethod
    def load(cls, path):
        with open(os.path.join(path, cls.filename), 'r') as fd:
            y = yaml.load(fd)

        project = cls(y['name'], y['stack'], deployparams=y['deployparams'])
        project.__dict__.update(y)
        return project

    @property
    def metadata(self):
        m = dict(type=self.__class__.__name__)
        m.update(self.__dict__)
        return m

    def sync_metadata(self, path):
        y = yaml.dump(self.metadata, default_flow_style=False)
        with open(os.path.join(path, Project.filename), 'w') as fd:
            fd.write(y)

    def init(self, force=False, update=False):
        self.stack.init(force=force, update=update)

    def deploy(self, force=False):

        if not self.is_deployed or force:
            self.stack.deploy(**self.deployparams)
        else:
            Console.info('Already deployed')

        self.is_deployed = True


class KWArgs(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __getitem__(self, key):
        return self.kwargs[key]

    def __setitem__(self, key, value):
        self.kwargs[key] = value


class BigDataStack(object):
    def __init__(self, dest,
                 repo='git://github.com/futuresystems/big-data-stack.git',
                 branch='master'):
        self.path = os.path.abspath(dest)
        self.repo = repo
        self.branch = branch
        self.local = os.path.isdir(repo)
        self._env = dict()

    def init(self, force=False, update=False):
        """Initialize by cloning (or updating if requested) a local copy of
        the Big Data Stack repository.

        :param bool force: setup previously setup project
        :type force: :class:`bool`
        :param bool update: update the local repository from the origin
        :type update: :class:`bool`

        """

        if not os.path.isdir(os.path.join(self.path, '.git')):
            Console.debug_msg('Cloning branch {} of {} to {}'
                              .format(self.branch, self.repo, self.path))
            Subprocess(['git', 'clone', '--recursive', '--branch',
                        self.branch, self.repo, self.path])

        elif update:
            Console.debug_msg('Updating to branch {} for {}'
                              .format(self.branch, self.path))
            Subprocess(['git', 'fetch', '--recurse-submodules', 'origin',
                        self.branch],
                       cwd=self.path)
            Subprocess(['git', 'checkout', self.branch], cwd=self.path)
            Subprocess(['git', 'merge', 'origin/{}'.format(self.branch)],
                       cwd=self.path)

        venvname = 'venv'
        venvdir = os.path.join(self.path, venvname)

        if force and os.path.isfile(os.path.join(venvdir, 'bin', 'activate')):
            Console.debug_msg('Removing {}'.format(venvdir))
            shutil.rmtree(venvdir)

        if not os.path.isdir(venvdir):
            Console.debug_msg('Creating virtualenv {}'.format(venvdir))
            Subprocess(['virtualenv', venvdir])

        self._env = get_virtualenv_environment(venvdir)
        cmd = ['pip', 'install', '-r', 'requirements.txt'] + (['-U']
                                                              if force
                                                              else [])
        Console.debug_msg('Installing requirements to {}'.format(venvdir))
        Subprocess(cmd, cwd=self.path, env=self._env)

    def deploy(self, ips=None, name=None, user=None, playbooks=None,
               defines=None, ping_max=10, ping_sleep=10):
        """Deploy the big-data-stack to a previously stood up cluster located
        at `ips` with login user `user`.

        :param ips: the ip addresses of the cluster to deploy to
        :type ips: :class:`list` of :class:`str` IP addresses
        :param name: the name of the cluster
        :type name: :class:`str`
        :param user: the login username of the cluster
        :type user: :class:`str`
        :param playbooks: the list of playbooks to deploy. These are paths relative to the root directory of the BDS repository.
        :type playbooks: :class:`list` of :class:`str`
        :param defines: the overridden variables defined for each playbook
        :type defines: :class:`dict` from playbook name to :class:`dict` of variable name to value
        :param ping_max: the maximum number of time to attempt to ping the cluster during the verification step.
        :type ping_max: :class:`int`
        :param ping_sleep: the number of seconds to wait between each attempt to ping
        :type ping_sleep: :class:`int`
        """
        assert ips is not None
        assert ping_max > 0, ping_max
        assert ping_sleep > 0, ping_sleep

        name = name or os.getenv('USER') + '-' + os.path.basename(self.path)
        user = user or 'defaultuser'
        playbooks = playbooks or list()
        defines = defines or defaultdict(list)

        Console.debug_msg('Calling mk-inventory in {}'.format(self.path))
        cmd = ['python', 'mk-inventory', '-n', name] + ips
        inventory = Subprocess(cmd, cwd=self.path, env=self._env)
        Console.debug_msg('Writing inventory file')
        Console.debug_msg('\n    ' + ('\n' + 4*' ').join(inventory.stdout.split('\n')))
        with open(os.path.join(self.path, 'inventory.txt'), 'w') as fd:
            fd.write(inventory.stdout)

        Console.info('Waiting for cluster to be accessible')
        ping_ok = False
        for i in xrange(ping_max):
            Console.debug_msg('Attempt {} / {}'.format(i+1, ping_max))
            try:
                Subprocess(['ansible', 'all', '-m', 'ping', '-u', user],
                           cwd=self.path, env=self._env,
                           stdout=None, stderr=None)
                ping_ok = True
                Console.debug_msg('Success!')
                break
            except SubprocessError:
                Console.debug_msg('Failure, sleeping for {} seconds'
                                  .format(ping_sleep))
                time.sleep(ping_sleep)

        if not ping_ok:
            msg = 'Ping Failure'
            reason = 'Unable to connect to all nodes'
            Console.error(' '.join([msg, reason]))
            raise SanityCheckError(message=msg, reason=reason)

        basic_command = ['ansible-playbook', '-u', user]
        Console.debug_msg('Running playbooks {}'.format(playbooks))
        for play in playbooks:
            cmd = basic_command + [play]
            define = ['{}={}'.format(k, v) for k, v in defines[play]]
            if define:
                cmd.extend(['-e', ','.join(define)])
            Console.info('Running playbook {} with overrides {}'
                         .format(play, define))
            Subprocess(cmd, cwd=self.path, env=self._env,
                       stdout=None, stderr=None)


class SanityCheckError(Exception):

    def __init__(self, message, reason):
        self.message = message
        self.reason = reason


class SanityChecker(object):

    def check_program(self, program):
        """Verify that the named program is available and executable by the user.

        :param program: commands that must be present by checking in the PATH
        :type programs: :class:`str`
        :returns: None
        :raises: :class:`SanityCheckError` on failure

        """

        PATH = os.getenv('PATH').split(':')

        for prefix in PATH:
            if os.path.exists(os.path.join(prefix, program)):
                return

        raise SanityCheckError('{} not installed correctly'.format(program),
                               '`{}` not found in $PATH'.format(program))

    def check_sshkey(self):
        """Ensure that the default ssh key exists with correct permissions

        :returns: None
        :raises: :class:`SanityCheckError`

        """

        valid_key_types = [
            'dsa',
            'ecdsa',
            'ed25519',
            'rsa',
            'rsa1'
        ]

        msg = 'SSH incorrectly configured'

        dotssh = os.path.expanduser('~/.ssh')

        if not os.path.exists(dotssh):
            raise SanityCheckError(
                msg,
                'directory `{}` does not exist, please run ssh-keygen'
                .format(dotssh))

        if not os.path.isdir(dotssh):
            raise SanityCheckError(msg,
                                   'path `{}` exists but is not a directory'
                                   .format(dotssh))

        s = os.stat(dotssh)
        if not bool(s.st_mode & stat.S_IRWXU):
            raise SanityCheckError(msg,
                                   'incorrect permissions, run: chmod 0700 "{}"'
                                   .format(dotssh))

        found = 0
        for typ in valid_key_types:
            private = os.path.join(dotssh, typ)
            public = private + '.pub'

            if os.path.exists(private) and os.path.exists(public):
                found += 1
                for p in [private, public]:
                    s = os.stat(p)
                    if bool(s.st_mode & (stat.S_IRGRP | stat.S_IROTH)):
                        raise SanityCheckError(
                            msg,
                            'Permissions on {} are too open')

        if found > 0:
            raise SanityCheckError(msg,
                                   'No id_<type> ssh keys found in {}'
                                   .format(dotssh))

    def check_github(self):
        """Ensure that the default ssh key has been uploaded to github.com

        :returns: None
        :raises: :class:`SanityCheckError` on failure
        """

        try:
            Subprocess(['ssh', '-T', 'git@github.com'])
        except SubprocessError as e:
            if "You've successfully authenticated" not in e.stderr:
                raise SanityCheckError(
                    'SSH authentication to github.com failed',
                    'did you add your public key to https://github.com/settings/ssh ?')


def sanity_check():
    """Verifies that the environment is set up correctly for BDS usage:

    :returns: boolean indicating pass or fail of the sanity check
    :rtype: :class:`bool`
    """

    def pprint_check(msg, maxchar=20, filler='.', stream=sys.stdout):
        """fancy way for printing out aligned messages

        :param msg: the message
        :param maxchar: maximum number of characters to elide
        :param filler: the filler character
        :param stream: the output stream
        """
        nfiller = maxchar - len(msg)
        nfiller = max(nfiller, 2)

        stream.write(msg)
        stream.write(nfiller * filler)

    checker = SanityChecker()

    # so all errors can be reported at once
    errors = []

    def check(fn, *args, **kws):
        try:
            fn(*args, **kws)
            print('OK')
        except SanityChecker as e:
            print('FAILED')
            errors.append(e)


    ################################################## programs

    # these should be available
    programs = [
        'python',
        'virtualenv',
        'pip',
        'git',
        'ssh',
    ]

    for prog in programs:
        pprint_check(prog)
        check(checker.check_program, prog)

    ################################################## ssh key

    pprint_check('ssh key')
    check(checker.check_sshkey)

    ################################################## ssh to github

    pprint_check('github')
    check(checker.check_github)


    ################################################## errors

    if len(errors) > 0:
        sys.stderr.write('The following errors were detected:\n\n')

        for e in errors:
            sys.stderr.write('* {}\n'.format(e.message))
            sys.stderr.write('  >{}\n'.format(e.reason))

        return False

    else:
        return True



from __future__ import print_function
import json
import os
import getpass
import socket

import pyaml
import time
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.cloud.vm import Vm
from cloudmesh_client.cloud.secgroup import SecGroup
from cloudmesh_client.cloud.network import Network
from cloudmesh_client.cloud.group import Group
from cloudmesh_client.cloud.counter import Counter
from cloudmesh_client.default import Default
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.ConfigDict import Username
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.common.Error import Error
from cloudmesh_client.common.LibcloudDict import LibcloudDict
from builtins import input
from cloudmesh_client.common.hostlist import Parameter
from cloudmesh_client.common.Shell import Shell
from pprint import pprint
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.cloud.image import Image
from cloudmesh_client.cloud.ip import Ip
from cloudmesh_client.common.util import search
from cloudmesh_client.db.CloudmeshDatabase import CloudmeshDatabase

class VmCommand(PluginCommand, CloudPluginCommand):
    topics = {"vm": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command vm")

    # noinspection PyUnusedLocal
    @command
    def do_vm(self, args, arguments):
        """
        ::

            Usage:
                vm default [--cloud=CLOUD][--format=FORMAT]
                vm refresh [all][--cloud=CLOUD]
                vm boot [--name=NAME]
                        [--cloud=CLOUD]
                        [--username=USERNAME]
                        [--image=IMAGE]
                        [--flavor=FLAVOR]
                        [--group=GROUP]
                        [--public]
                        [--secgroup=SECGROUP]
                        [--key=KEY]
                        [--dryrun]
                vm boot [--n=COUNT]
                        [--cloud=CLOUD]
                        [--username=USERNAME]
                        [--image=IMAGE]
                        [--flavor=FLAVOR]
                        [--group=GROUP]
                        [--public]
                        [--secgroup=SECGROUP]
                        [--key=KEY]
                        [--dryrun]
                vm ping [NAME] [N]
                vm console [NAME]
                         [--group=GROUP]
                         [--cloud=CLOUD]
                         [--force]
                vm start [NAMES]
                         [--group=GROUP]
                         [--cloud=CLOUD]
                         [--force]
                vm stop [NAMES]
                        [--group=GROUP]
                        [--cloud=CLOUD]
                        [--force]
                vm terminate [NAMES]
                          [--group=GROUP]
                          [--cloud=CLOUD]
                          [--force]
                vm delete [NAMES]
                          [--group=GROUP]
                          [--cloud=CLOUD]
                          [--keep]
                          [--dryrun]
                vm ip assign [NAMES]
                          [--cloud=CLOUD]
                vm ip show [NAMES]
                           [--group=GROUP]
                           [--cloud=CLOUD]
                           [--format=FORMAT]
                           [--refresh]
                vm ip inventory [NAMES]
                                [--header=HEADER]
                                [--file=FILE]
                vm ssh [NAME] [--username=USER]
                         [--quiet]
                         [--ip=IP]
                         [--cloud=CLOUD]
                         [--key=KEY]
                         [--command=COMMAND]
                         [--modify-knownhosts]
                vm rename [OLDNAMES] [NEWNAMES] [--force] [--dryrun]
                vm list [NAMES]
                        [--cloud=CLOUDS|--active]
                        [--group=GROUP]
                        [--format=FORMAT]
                        [--refresh]
                vm status [NAMES]
                vm wait [--cloud=CLOUD] [--interval=SECONDS]
                vm info [--cloud=CLOUD]
                        [--format=FORMAT]
                vm check NAME
                vm username USERNAME [NAMES] [--cloud=CLOUD]

            Arguments:
                COMMAND        positional arguments, the commands you want to
                               execute on the server(e.g. ls -a) separated by ';',
                               you will get a return of executing result instead of login to
                               the server, note that type in -- is suggested before
                               you input the commands
                NAME           server name. By default it is set to the name of last vm from database.
                NAMES          server name. By default it is set to the name of last vm from database.
                KEYPAIR_NAME   Name of the openstack keypair to be used to create VM. Note this is
                               not a path to key.
                NEWNAMES       New names of the VM while renaming.
                OLDNAMES       Old names of the VM while renaming.

            Options:
              -H --modify-knownhosts  Do not modify ~/.ssh/known_hosts file when ssh'ing into a machine
                --username=USERNAME  the username to login into the vm. If not specified it will be guessed
                                     from the image name and the cloud
                --ip=IP          give the public ip of the server
                --cloud=CLOUD    give a cloud to work on, if not given, selected
                                 or default cloud will be used
                --count=COUNT    give the number of servers to start
                --detail         for table print format, a brief version
                                 is used as default, use this flag to print
                                 detailed table
                --flavor=FLAVOR  give the name or id of the flavor
                --group=GROUP          give the group name of server
                --secgroup=SECGROUP    security group name for the server
                --image=IMAGE    give the name or id of the image
                --key=KEY        specify a key to use, input a string which
                                 is the full path to the private key file
                --keypair_name=KEYPAIR_NAME   Name of the openstack keypair to be used to create VM.
                                              Note this is not a path to key.
                --user=USER      give the user name of the server that you want
                                 to use to login
                --name=NAME      give the name of the virtual machine
                --force          rename/ delete vms without user's confirmation
                --command=COMMAND
                                 specify the commands to be executed


            Description:
                commands used to boot, start or delete servers of a cloud

                vm default [options...]
                    Displays default parameters that are set for vm boot either on the
                    default cloud or the specified cloud.

                vm boot [options...]
                    Boots servers on a cloud, user may specify flavor, image .etc, otherwise default values
                    will be used, see how to set default values of a cloud: cloud help

                vm start [options...]
                    Starts a suspended or stopped vm instance.

                vm stop [options...]
                    Stops a vm instance .

                vm delete [options...]
                    Delete servers of a cloud, user may delete a server by its name or id, delete servers
                    of a group or servers of a cloud, give prefix and/or range to find servers by their names.
                    Or user may specify more options to narrow the search

                vm floating_ip_assign [options...]
                    assign a public ip to a VM of a cloud

                vm ip show [options...]
                    show the ips of VMs

                vm ssh [options...]
                    login to a server or execute commands on it

                vm list [options...]
                    same as command "list vm", please refer to it

                vm status [options...]
                    Retrieves status of last VM booted on cloud and displays it.

            Tip:
                give the VM name, but in a hostlist style, which is very
                convenient when you need a range of VMs e.g. sample[1-3]
                => ['sample1', 'sample2', 'sample3']
                sample[1-3,18] => ['sample1', 'sample2', 'sample3', 'sample18']

            Quoting commands:
                cm vm login gvonlasz-004 --command=\"uname -a\"
        """

        """

        # terminate
        #  issues a termination to the cloud, keeps vm in database

        # delete
        #   issues a terminate if not already done
        #      (remember you do not have to go to cloud if state is already terminated)
        #   deletes the vm from database
        #


        # bulk rename

        rename abc[0-1] def[3-4]       renames the abc0,abc1 -> def3,def4

        if arguments["rename"]:
            oldnames = Parameter.expand(arguments["OLDNAME"])
            newnames = Parameter.expand(arguments["NEWNAME"])

            # check if new names ar not already taken
            # to be implemented

            if len(oldnames) == len(newnames):
                for i in range(0, len(oldnames)):
                    oldname = oldnames[i]
                    newname = newnames[i]
                if newname is None or newname == '':
                    print("New node name cannot be empty")
                else:
                    print(Cluster.rename_node(clusterid, oldname, newname))
        """

        cm = CloudmeshDatabase()

        def _print_dict(d, header=None, output='table'):
            return Printer.write(d, order=["id", "name", "status"], output=output, sort_keys=True)

        def _print_dict_ip(d, header=None, output='table'):
            return Printer.write(d, order=["network", "version", "addr"], output=output, sort_keys=True)

        def _refresh_cloud(cloud):
            try:
                msg = "Refresh VMs for cloud {:}.".format(cloud)
                if Vm.refresh(cloud=cloud):
                    Console.ok("{:} OK.".format(msg))
                else:
                    Console.error("{:} failed".format(msg), traceflag=False)
            except Exception as e:
                Console.error("Problem running VM refresh", traceflag=False)

        def _get_vm_names():

            vm_list  = cm.find(kind="vm")

            vms = [vm["name"] for vm in vm_list]

            names = pattern = arguments["NAMES"]
            if pattern is not None:
                if "*" in pattern:
                    names = search(vms, pattern)
                else:
                    names = Parameter.expand(names)

            if names == ['last'] or names is None:
                names == [Default.vm]

            return vm_list, names

        cloud = arguments["--cloud"] or Default.cloud

        config = ConfigDict("cloudmesh.yaml")
        active_clouds = config["cloudmesh"]["active"]

        def _refresh(cloud):
            all = arguments["all"] or None

            if all is None:

                _refresh_cloud(cloud)
            else:
                for cloud in active_clouds:
                    _refresh_cloud(cloud)

        arg = dotdict(arguments)

        arg.cloud = arguments["--cloud"] or Default.cloud
        arg.image = arguments["--image"] or Default.get(name="image", category=arg.cloud)
        arg.flavor = arguments["--flavor"] or Default.get(name="flavor", category=arg.cloud)
        arg.group = arguments["--group"] or Default.group
        arg.secgroup = arguments["--secgroup"] or Default.secgroup
        arg.key = arguments["--key"] or Default.key
        arg.dryrun = arguments["--dryrun"]
        arg.name = arguments["--name"]
        arg.format = arguments["--format"] or 'table'
        arg.refresh = Default.refresh or arguments["--refresh"]
        arg.count = int(arguments["--n"] or 1)
        arg.dryrun = arguments["--dryrun"]
        arg.verbose = not arguments["--quiet"]

        #
        # in many cases use NAMES
        # if arg.NAMES is not None:
        #   arg.names = Parameter.expand(arg.NAMES)   #  gvonlasz[001-002]  gives ["gvonlasz-001", "gvonlasz-002"]
        # else:
        #    arg.names = None
        #

        if arguments["boot"]:

            arg.username = arguments["--username"] or Image.guess_username_from_category(arg.cloud, arg.image)
            is_name_provided = arg.name is not None

            arg.user = Default.user

            for index in range(0, arg.count):
                vm_details = dotdict({
                    "cloud": arg.cloud,
                    "name": Vm.generate_vm_name(),
                    "image": arg.image,
                    "flavor": arg.flavor,
                    "key": arg.key,
                    "secgroup": arg.secgroup,
                    "group": arg.group,
                    "username": arg.username,
                    "user": arg.user
                })

                try:

                    if arg.dryrun:
                        print(Printer.attribute(vm_details, output=arg.format))
                        msg = "dryrun info. OK."
                        Console.ok(msg)
                    else:
                        vm_id = Vm.boot(**vm_details)

                        if vm_id is None:
                            msg = "info. failed."
                            Console.error(msg, traceflag=False)
                            return ""

                        # set name and counter in defaults
                        Default.set_vm(value=vm_details.name)
                        if is_name_provided is False:
                            Default.incr_counter("name")

                        # Add to group
                        if vm_id is not None:
                            Group.add(name=vm_details.group,
                                      species="vm",
                                      member=vm_details.name,
                                      category=vm_details.cloud)

                        msg = "info. OK."
                        Console.ok(msg)

                except Exception as e:
                    # error -> warning: probably vm launched with minor errors
                    Console.warning("Problem booting instance {name}".format(**vm_details))
                    Console.info("%s" % e)
                    Console.error("Problem booting instance {name}".format(**vm_details), traceflag=False)

        elif arguments["username"]:

            arg.username = arguments["--username"] or Image.guess_username(arg.image)

            cloud = arg.cloud
            username = arg.USERNAME
            if arg.NAMES is None:
                names = [Default.vm]
            else:
                names = Parameter.expand(arg.NAMES)

            if len(names) == 0:
                return

            for name in names:
                arg.name = name
                Console.ok("Set username for {cloud}:{name} to {USERNAME}".format(**arg))
                Vm.set_login_user(name=name, cloud=cloud, username=username)

        elif arguments["default"]:
            try:
                count = Default.get_counter()
                prefix = Username()

                if prefix is None or count is None:
                    Console.error("Prefix and Count could not be retrieved correctly.", traceflag=False)
                    return

                vm_name = prefix + "-" + str(count).zfill(3)
                arg = {
                    "name": vm_name,
                    "cloud": arguments["--cloud"] or Default.cloud
                }
                for attribute in ["image", "flavor"]:
                    arg[attribute] = Default.get(name=attribute, category=cloud)
                for attribute in ["key", "group", "secgroup"]:
                    arg[attribute] = Default.get(name=attribute, category='general')

                output = arguments["--format"] or "table"
                print(Printer.attribute(arg, output=output))
                msg = "info. OK."
                Console.ok(msg)
                ValueError("default command not implemented properly. Upon "
                           "first install the defaults should be read from yaml.")
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem listing defaults", traceflag=False)

        elif arguments["ping"]:
            try:
                if arguments["NAME"] is None and arguments["N"] is None:
                    name = arguments["NAME"] or Default.vm
                    n = arguments["N"] or 1
                elif arguments["NAME"].isdigit():
                    n = arguments["NAME"]
                    name = Default.vm
                else:
                    name = arguments["NAME"] or Default.vm
                    n = arguments["N"] or 1

                print("Ping:", name, str(n))

                vm = dotdict(Vm.list(name=name, category=cloud, output="dict")["dict"])

                ip = vm.floating_ip

                result = Shell.ping(host=ip, count=n)
                print(result)

            except Exception as e:
                Console.error(e.message, traceflag=False)

        elif arguments["console"]:
            try:
                name = arguments["NAME"] or Default.vm

                vm = dotdict(Vm.list(name=name, category=cloud, output="dict")["dict"])

                cloud_provider = CloudProvider(cloud).provider
                vm_list = cloud_provider.list_console(vm.uuid)
                print(vm_list)
                msg = "info. OK."
                Console.ok(msg)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem retrieving status of the VM", traceflag=False)

        elif arguments["status"]:
            try:
                cloud_provider = CloudProvider(cloud).provider
                vm_list = cloud_provider.list_vm(cloud)

                vms = [vm_list[i]["name"] for i in vm_list ]
                print ("V", vms)

                pattern = arguments["NAMES"]
                if pattern is not None:
                    if "*" in pattern:
                        print ("serach")
                        names  = search(vms, pattern)
                    else:
                        names = Parameter.expand()
                    for i in vm_list:
                        if vm_list[i]["name"] in names:
                            print("{} {}".format(vm_list[i]["status"], vm_list[i]["name"]))
                else:
                    print("{} {}".format(vm_list[0]["status"], vm_list[0]["name"]))
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem retrieving status of the VM", traceflag=True)
        elif arguments["wait"]:
            interval = arguments["--interval"] or 5
            try:
                cloud_provider = CloudProvider(cloud).provider
                for i in range(1,10):
                    vm_list = cloud_provider.list_vm(cloud)
                    time.sleep(float(1))
                    d = {}
                    for id in vm_list:
                        vm = vm_list[id]
                        d[vm["name"]] = vm["status"]
                    print (d)
                    print("{} {}".format(vm_list[0]["status"], vm_list[0]["name"]))
                    if vm_list[0]["status"] in ['ACTIVE']:
                        return
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem retrieving status of the VM", traceflag=True)

        elif arguments["info"]:
            try:
                cloud_provider = CloudProvider(cloud).provider
                vms = cloud_provider.list_vm(cloud)
                vm = vms[0]
                output_format = arguments["--format"] or "table"
                print(Printer.attribute(vm, output=output_format))
                msg = "info. OK."
                Console.ok(msg)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem retrieving status of the VM", traceflag=False)

        elif arguments["check"]:

            test = {}
            try:

                names = Parameter.expand(arguments["NAME"])
                id = 0
                for name in names:
                    print("Not implemented: {}".format(name))
                    # TODO: check the status of the vms
                    status = "active"
                    # TODO: check if they have a floating ip
                    # TODO: get ip
                    floating_ip = "127.0.0.1"
                    ip = True
                    # ping
                    # TODO: ping the machine with the shell command
                    ping = True
                    # check if one can login and run a command
                    check = False
                    try:
                        r = Shell.execute("uname", "-a")
                        # do a real check
                        check = True
                    except:
                        check = False
                    test[name] = {
                        "id": id,
                        "name": name,
                        "status": status,
                        "ip": ip,
                        "ping": ping,
                        "login": check
                    }
                    id += 1

                pprint(test)

                print(Printer.write(test,
                                    order=["id",
                                           "name",
                                           "status",
                                           "ip",
                                           "ping",
                                           "login"],
                                    output="table",
                                    sort_keys=True))

                msg = "not yet implemented. failed."
                Console.error(msg, traceflag=False)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem retrieving status of the VM", traceflag=False)

        elif arguments["start"]:
            try:
                servers = Parameter.expand(arguments["NAMES"])

                # If names not provided, take the last vm from DB.
                if len(servers) == 0:
                    last_vm = Default.vm
                    if last_vm is None:
                        Console.error("No VM records in database. Please run vm refresh.", traceflag=False)
                        return ""
                    name = last_vm["name"]
                    # print(name)
                    servers = list()
                    servers.append(name)

                group = arguments["--group"]
                force = arguments["--force"]

                # if default cloud not set, return error
                if not cloud:
                    Console.error("Default cloud not set.", traceflag=False)
                    return ""

                Vm.start(cloud=cloud, servers=servers)

                msg = "info. OK."
                Console.ok(msg)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem starting instances", traceflag=False)

        elif arguments["stop"]:
            try:
                servers = Parameter.expand(arguments["NAMES"])

                # If names not provided, take the last vm from DB.
                if servers is None or len(servers) == 0:
                    last_vm = Default.vm
                    if last_vm is None:
                        Console.error("No VM records in database. Please run vm refresh.", traceflag=False)
                        return ""
                    name = last_vm["name"]
                    # print(name)
                    servers = list()
                    servers.append(name)

                group = arguments["--group"]
                force = arguments["--force"]

                # if default cloud not set, return error
                if not cloud:
                    Console.error("Default cloud not set.", traceflag=False)
                    return ""

                Vm.stop(cloud=cloud, servers=servers)

                msg = "info. OK."
                Console.ok(msg)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem stopping instances", traceflag=False)

        elif arguments["refresh"]:

            _refresh(cloud)

        elif arguments["delete"]:

            dryrun = arguments["--dryrun"]
            group = arguments["--group"]
            force = not arguments["--keep"]
            cloud = arguments["--cloud"]
            vms, servers = _get_vm_names()

            if servers in [None, []]:
                Console.error("No vms found.", traceflag=False)
                return ""

            for server in servers:
                if dryrun:
                    Console.ok("Dryrun: delete {}".format(server))
                else:
                   Vm.delete(servers=[server], force=force)

            return ""

        elif arguments["ip"] and arguments["assign"]:
            if arguments["NAMES"] is None:
                names = [Default.vm]
            else:
                names = Parameter.expand(arguments["NAMES"])

            for name in names:

                # ip = Network.get_floatingip(....)

                vm = dotdict(Vm.list(name=name, category=cloud, output="dict")["dict"])

                if vm.floating_ip is None:

                    Console.ok("Assign IP to {}".format(name))

                    try:
                        floating_ip = Network.find_assign_floating_ip(cloudname=cloud,
                                                                      instance_id=name)

                        Vm.refresh(cloud=cloud)

                        if floating_ip is not None:
                            print(
                                "Floating IP assigned to {:} is {:}".format(
                                    name, floating_ip))
                            msg = "info. OK."
                            Console.ok(msg)
                    except Exception as e:

                        Console.error("Problem assigning floating ips.", traceflag=False)

                else:
                    Console.error("VM {} already has a floating ip: {}".format(name, vm.floating_ip), traceflag=False)


        elif arguments["ip"] and arguments["inventory"]:

            vms, names = _get_vm_names()

            if names in [None, []]:
                if str(Default.vm) in ['None', None]:
                    Console.error("The default vm is not set.", traceflag=False)
                    return ""
                else:
                    names = [Default.vm]

            header = arguments["--header"] or "[servers]"
            filename = arguments["--file"] or "inventory.txt"

            try:
                vm_ips = []
                for vm in vms:
                    if  vm["name"] in names:
                        print (vm["name"])
                        vm_ips.append(vm["floating_ip"])

                result = header + "\n"

                result += '\n'.join(vm_ips)
                Console.ok("Creating inventory file: {}".format(filename))

                Console.ok(result)

                with open(filename, 'w') as f:
                    f.write(result)



            except Exception as e:
                Console.error("Problem getting ip addresses for instance", traceflag=True)

        elif arguments["ip"] and arguments["show"]:
            if arguments["NAMES"] is None:
                if str(Default.vm) in ['None', None]:
                    Console.error("The default vm is not set.", traceflag=False)
                    return ""
                else:
                    names = [Default.vm]
            else:
                names = Parameter.expand(arguments["NAMES"])

            group = arguments["--group"]
            output_format = arguments["--format"] or "table"
            refresh = arguments["--refresh"]

            try:

                ips = Ip.list(cloud=arg.cloud, output=output_format, names=names)
                print(ips)
            except Exception as e:
                Console.error("Problem getting ip addresses for instance", traceflag=False)

        elif arguments["ssh"]:

            def _print(msg):
                if arg.verbose:
                    Console.msg(msg)

            chameleon = "chameleon" in ConfigDict(filename="cloudmesh.yaml")["cloudmesh"]["clouds"][arg.cloud][
                "cm_host"]

            # TODO
            # username should be fetched from --username parameter or from definition in yaml file in any cloud
            if chameleon:
                arg.username = "cc"
            elif arg.cloud == "azure":
                arg.username = arguments["--username"] or ConfigDict(filename="cloudmesh.yaml")["cloudmesh"]["clouds"]["azure"]["default"]["username"]
            else:
                if arg.username is None:
                    Console.error("Could not guess the username of the vm", traceflag=False)
                    return
                arg.username = arguments["--username"] or Image.guess_username(arg.image)
            arg.command = arguments["--command"]

            data = dotdict({
                'name': arguments["NAME"] or Default.vm,
                'username': arg.username,
                'cloud': arg.cloud,
                'command': arg.command
            })

            _print("login {cloud}:{username}@{name}".format(**data))

            vm = Vm.get(data.name, category=data.cloud)


            Vm.set_login_user(name=data.name, cloud=data.cloud, username=data.username)

            data.floating_ip = vm.floating_ip
            data.key = arguments["--key"] or Default.key

            _print(Printer.attribute(data))

            '''
            if vm.username is None:
                user_from_db = Vm.get_login_user(vm.name, vm.cloud)

                user_suggest = user_from_db or Default.user
                username = input("Username (Default: {}):".format(user_suggest)) or user_suggest

            Vm.set_login_user(name=data.name, cloud=cloud, username=data.username)
            '''

            ip = arguments["--ip"]
            commands = arguments["--command"]

            ip_addresses = []

            cloud_provider = CloudProvider(cloud).provider
            ip_addr = cloud_provider.get_ips(vm.name)

            ipaddr_dict = Vm.construct_ip_dict(ip_addr, cloud)
            for entry in ipaddr_dict:
                ip_addresses.append(ipaddr_dict[entry]["addr"])

            if len(ip_addresses) > 0:
                if ip is not None:
                    if ip not in ip_addresses:
                        Console.error("IP Address specified does not match with the host.", traceflag=False)
                        return ""
                else:
                    _print("Determining IP Address to use with a ping test.")
                    # This part assumes that the ping is allowed to the machine.
                    for ipadd in ip_addresses:
                        _print("Checking {:}...".format(ipadd))
                        try:
                            # Evading ping test, as ping is not enabled for VMs on Azure cloud
                            # socket.gethostbyaddr(ipadd)
                            # ip will be set if above command is successful.
                            ip = ipadd
                        except socket.herror:
                            _print("Cannot reach {:}.".format(ipadd))

                if ip is None:
                    _print("Unable to connect to the machine")
                    return ""
                else:
                    _print("IP to be used is: {:}".format(ip))

                #
                # TODO: is this correctly implemented
                #
                if not cloud == 'azure':
                    SecGroup.enable_ssh(cloud=cloud)

                if arg.verbose:
                    Console.info("Connecting to Instance at IP:" + format(ip))
                # Constructing the ssh command to connect to the machine.
                sshcommand = "ssh"
                if arg.key is not None:
                    sshcommand += " -i {:}".format(arg.key)
                sshcommand += " -o StrictHostKeyChecking=no"
                sshcommand += '' if arguments['--modify-knownhosts'] else ' -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no'
                sshcommand += " {:}@{:}".format(data.username, ip)
                if commands is not None:
                    sshcommand += " \"{:}\"".format(commands)

                # print(sshcommand)
                os.system(sshcommand)
            else:
                Console.error("No Public IPs found for the instance", traceflag=False)

        elif arguments["list"]:

            # groups = Group.list(output="dict")

            arg = dotdict(arguments)
            arg.names = arguments["NAMES"]

            arg.group = arguments["--group"]
            if arg.group is None:
                arg.group = []
            else:
                arg.group = Parameter.expand(arguments["--group"])

            arg.refresh = arguments["--refresh"] or Default.refresh

            if arg.NAMES is not None:
                arg.names = Parameter.expand(arguments["NAMES"])
            else:
                arg.names = ["all"]

            _format = arguments["--format"] or "table"

            if arguments["--active"]:
                clouds = active_clouds
            else:
                if arguments["--cloud"]:
                    clouds = Parameter.expand(arguments["--cloud"])
                else:
                    clouds = [Default.cloud]

            try:

                d = ConfigDict("cloudmesh.yaml")
                for cloud in clouds:

                    if arg.refresh:
                        _refresh(cloud)

                    Console.ok("Listing VMs on Cloud: {:}".format(cloud))

                    vms = Vm.list(category=cloud, output="raw")

                    # print ("XXX", type(vms), vms)

                    if vms is None:
                        break

                    result = []
                    if "all" in arg.names:
                        if result is None:
                            result = []
                        else:
                            result = vms
                    elif arg.group is not None and len(arg.group) > 0:
                        for vm in vms:
                            if vm["group"] in arg.group:
                                result.append(vm)
                    elif arg.names is not None and len(arg.names) > 0:
                        for vm in vms:
                            if vm["name"] in arg.names:
                                result.append(vm)

                    if len(result) > 0:
                        # print(result)
                        (order, header) = CloudProvider(cloud).get_attributes("vm")
                        print(Printer.write(result,
                                            order=order,
                                            output=_format)
                              )
                    else:
                        Console.error("No data found with requested parameters.", traceflag=False)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem listing all instances", traceflag=False)


        elif arguments["rename"]:
            try:
                oldnames = Parameter.expand(arguments["OLDNAMES"])
                newnames = Parameter.expand(arguments["NEWNAMES"])
                force = arguments["--force"]

                if oldnames is None or newnames is None:
                    Console.error("Wrong VMs specified for rename", traceflag=False)
                elif len(oldnames) != len(newnames):
                    Console.error("The number of VMs to be renamed is wrong",
                                  traceflat=False)
                else:
                    for i in range(0, len(oldnames)):
                        oldname = oldnames[i]
                        newname = newnames[i]
                        if arguments["--dryrun"]:
                            Console.ok("Rename {} to {}".format(oldname, newname))
                        else:
                            Vm.rename(cloud=cloud,
                                      oldname=oldname,
                                      newname=newname,
                                      force=force
                                      )
                    msg = "info. OK."
                    Console.ok(msg)
            except Exception as e:
                # Error.traceback(e)
                Console.error("Problem deleting instances", traceflag=False)

        return ""



from __future__ import print_function

import os
import sys

# import pip
import yaml
from setuptools.command.install import install

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.util import banner

make = """
    test:
    - echo "hallo world"
    github:
    - git commit -a
    - git push
    clean:
    - rm -rf *.egg-info *.eggs
    - rm -rf docs/build
    - rm -rf build
    - rm -rf dist
    doc:
    - sphinx-apidoc -o docs/source cloudmesh_client
    - cd docs; make -f Makefile html
    view:
    - open docs/build/html/index.html
    pypi:
    - python setup.py install
    - python setup.py sdist bdist_wheel
    - python setup.py bdist_wheel upload -r {repo}
    - python setup.py sdist upload -r {repo}
    register:
    - python setup.py register -r {repo}
    tag:
    - bin/new_version.sh
    rmtag:
    - git tag
    - echo "rm Tag?"; read TAG; git tag -d $$TAG; git push origin :refs/tags/$$TAG
    install:
    - python setup.py install
"""

clean_python = '''
    find . -name "*~" -exec rm \{\} \;
    find . -name "*.pyc" -exec rm \{\} \;
'''


def parse_requirements(filename):
    """
    load the requirement form the specified file
    :param filename: the filename
    :return:
    """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


def os_execute(commands):
    for command in commands.split("\n"):
        command = command.strip()
        print(command)
        os.system(command)


def get_version_from_git():
    r = Shell.git('tag').split("\n")[-1]
    return r


def check_pip():
    '''
    major = int(pip.__version__.split(".")[0])
    if major < 7:
        print("")
        print("    ERROR: Pip version", pip.__version__, "is to old.")
        print("    Tip:   Please update pip with ")
        print("")
        print("             pip install pip -U")
        print("")
        sys.exit()
    '''
    pass

def makefile(tag, **kwargs):
    script = "\n".join(yaml.load(make)[tag])
    commands = "\n".join(yaml.load(make)[tag]).format(**kwargs)

    banner("RUNNING")
    os_execute(commands)


def Make(action, **kwargs):

    class InstallAction(install):
        def run(self):
            makefile(action, **kwargs)
            if action == "clean":
                os.system(clean_python)

    return InstallAction





""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_cloud/test_scripts.py

nosetests -v --nocapture tests/test_scripts.py

or

nosetests -v tests/test_scripts.py

"""

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from cloudmesh_client.default import Default


# noinspection PyPep8Naming
class Test_script:
    """tests script command"""

    data = dotdict({
        "cloud": Default.cloud,
        "group": "mygroup"
    })

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return str(result)

    def setup(self):
        self.scripts = [
            ("bash.cm", "README.rst"),
            ("comment.cm", "/* test"),
            ("var.cm", "username"),
            ("py.cm", "2"),
            ("terminal.cm", ""),
            # ("hpc.cm", "bravo"),
            ("key.cm", ""),
            ("reservedemo.cm", "cloudnauts"),
            # ("cloud.cm", "Active"),
            # ("nova.cm", "Status"),
            # ("demo.cm", "ephemeral_disk"),

            # BROKEN:
            # ("group.cm", ""),
            # ("sync.cm", ""),
            # ("secgroup.cm", ""),
            # ("cluster.cm", ""),
            # ("vm.cm", ""),
            # ("network.cm", ""),

        ]
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_001(self):
        HEADING("cm script set india")
        for self.data["script"], self.data["check"] in self.scripts:
            result = self.run("cm scripts/{script}")
            assert self.data["check"] in result



""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_cloud/test_keys.py:Test_keys.test_001

nosetests -v --nocapture

or

nosetests -v

"""
from __future__ import print_function

from pprint import pprint

import yaml
from cloudmesh_client.cloud.key import Key

from cloudmesh_client.cloud.iaas.provider.openstack.CloudProviderOpenstackAPI import CloudProviderOpenstackAPI
from cloudmesh_client import CloudmeshDatabase
from cloudmesh_client import ConfigDict
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.SSHkey import SSHkey
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from cloudmesh_client.default import Default


# noinspection PyMethodMayBeStatic,PyMethodMayBeStatic,PyPep8Naming
class Test_keys:
    cm = CloudmeshDatabase()

    # noinspection PyTypeChecker
    data = dotdict({
        "cloud": Default.cloud,
        "username": ConfigDict("cloudmesh.yaml")["cloudmesh"]["github"]["username"],
        "key": Default.user
    }
    )

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return result

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def clean_db(self):
        """create new db"""
        print("clean db")
        self.cm.clean()
        result = self.run("cm default cloud={cloud}")
        result = self.run("cm default cloud")
        assert self.data.cloud in result

    def test_001(self):
        HEADING("clen")
        self.cm.clean()

    def test_002(self):
        HEADING("reading the keys from ~/.ssh")

        banner("ssh keys")
        Key.get_from_dir("~/.ssh")
        mykeys = Key.all(output="dict")

        assert len(mykeys) > 0

    def test_003(self):
        HEADING("reading the keys from github")

        self.cm.clean()

        config = ConfigDict("cloudmesh.yaml")
        git_username = config['cloudmesh']['github']['username']

        banner("git hub")
        Key.get_from_git(git_username)

        mykeys = Key.all(output="dict")
        count1 = len(mykeys)

        banner("all")
        key = Key.get(git_username, output="dict")
        print(key)

        mykeys = Key.all(output="dict")
        count2 = len(mykeys)

        print(mykeys)

        Key.delete(name='github-0')

        mykeys = Key.all(output="table")

        print(mykeys)

        assert count1 > 0 and count2 > 0

    def test_004(self):
        HEADING("testing properties in SSHKey")

        sshkey = SSHkey("~/.ssh/id_rsa.pub")
        pprint(sshkey.key)
        print("Fingerprint:", sshkey.fingerprint)
        pprint(sshkey.__key__)
        print("sshkey", sshkey)
        print("str", str(sshkey))
        print(sshkey.type)
        assert True

    def test_005(self):
        HEADING()

        Key.delete()
        Key.add_from_path("~/.ssh/id_rsa.pub")

        d = Key.all(output="dict")
        print(d)

        print(Printer.write(d, output="table"))
        assert 'id_rsa.pub' in str(d)

        d = Key.find(name='rsa')
        print('find function: ', d)

        Key.delete(name='rsa')

        d = Key.all(output="dict")

        assert d is None

    def test_006(self):
        """
        cm key add testkey ~/.ssh/id_rsa.pub
        """
        HEADING()
        self.clean_db()
        result = self.run("cm key add {key} --source=~/.ssh/id_rsa.pub")
        result = self.run("cm key list")

        assert "{key}".format(**self.data) in result
        assert "file:" in result

    def test_007(self):
        """
        cm key add --git --name=testkey ~/.ssh/id_rsa.pub
        """
        HEADING()
        self.clean_db()

        result = self.run("cm key add --git ")
        result = self.run("cm key list")

        username = "{username}_git".format(**self.data)

        assert username in result

    def test_008(self):
        """
        key add --ssh --name=testkey
        """
        HEADING()
        self.clean_db()
        result = self.run("cm key add {key} --ssh")
        result = self.run("cm key list")
        assert "{key}".format(**self.data) in result
        assert "ssh" in result

    def test_009(self):
        """
        cm key list --source=db
        """

        HEADING()
        self.clean_db()
        result = self.run("cm key add {key} --ssh")
        result = self.run("cm key list --source=db")

        assert "{key}".format(**self.data) in result
        assert "ssh" in result

    def test_010(self):
        """
        cm key list --source=db --format=json
        """

        HEADING()
        result = self.run("cm key list --source=db --format=json")
        print(result)

        assert "{" in result
        assert "{key}".format(**self.data) in result

    def test_011(self):
        """
        cm key list --source=db --format=yaml
        """

        HEADING()
        result = self.run("cm key list --source=db --format=yaml")

        print("--------")
        print(result)
        print("--------")

        # d = yaml.load(str(result))


        # assert "testkey" in result
        assert ":" in str(result)

    def test_012(self):
        """
        cm key list --source=git
        """

        HEADING()
        result = self.run("cm key list --source=git")
        assert "_git_" in result

    def test_013(self):
        """
        cm key list --source=cloudmesh
        """

        HEADING()
        result = self.run("cm key list --source=yaml")
        print(result)
        assert ":" in result

    def test_014(self):
        """
        cm key list --source=ssh
        """

        HEADING()
        result = self.run("cm key list --source=ssh")
        assert "rsa" in result

    def test_015(self):
        """
        cm key get testkey
        """

        HEADING()
        result = self.run("cm key get {key}")
        assert "{key}".format(**self.data) in result

    def test_016(self):
        """
        cm key default testkey
        """

        HEADING()
        result = self.run("cm default key={key}")
        assert "ok." in result

        result = self.run("cm default key")
        assert "{key}".format(**self.data) in result

    def test_017(self):
        """
        cm key delete testkey
        """

        HEADING()
        self.clean_db()
        result = self.run("cm key add {key} --ssh")
        result = self.run("cm key delete {key}")
        assert "{key}".format(**self.data) in result
        assert "OK." in result

        result = self.run("cm key list")

        assert "None" in str(result)

    def test_018(self):
        """
        cm key delete --all
        """

        HEADING()
        result = self.run("cm key add {key} --ssh")
        result = self.run("cm key delete --all")
        assert "OK." in result

    def test_019(self):
        HEADING()
        self.clean_db()
        result = self.run("cm key add {key} --ssh")
        result = self.run("cm key upload {key}")

        cloudname = self.data.cloud
        d = ConfigDict("cloudmesh.yaml")
        cloud_details = d["cloudmesh"]["clouds"][cloudname]

        cloud = CloudProviderOpenstackAPI(cloudname, cloud_details)

        d = cloud.list_key(cloudname)
        pprint(d)

        assert len(d) > 0
        result = self.run("cm key delete {key} --cloud={cloud}")
        assert 'OK' in result

    def test_020(self):
        HEADING("set key to user defaults")

        result = self.run("cm key add {} --ssh".format(Default.user))
        result = self.run("cm default key={}".format(Default.user))



    # def test_021(self):
    #    HEADING()
    #    print(Key.get_from_yaml())

    # def test_022(self):
    #    """
    #    cm key list --source=cloudmesh
    #    """
    #    HEADING()
    #    result = self.run("cm key list --source=yaml")
    #    assert "ssh" in result
    #    assert "fingerprint" in result



from __future__ import print_function
from cloudmesh_client.extend.my import my_i, my_c
from cloudmesh_client.common.DynamicClass import DynamicClass

def z_a(args):
    print ("z_a", args)

def z_c(cls, args):
    print ("z_c", args)

def z_b(cls, args):
    print ("z_b", args)


class A(DynamicClass):
    pass

if __name__ == "__main__":
    a = A()
    a.add_instance_method(my_i)
    a.my_i("i am my_i")

    A.add_classmethod(my_c)
    a.my_c("i am my_c")

    A.load_classmethod("cloudmesh_client.extend.my.my_cc")
    a.my_cc("i am my_cc")

    b = A()

    b.my_cc("i am my_cc")

    A.z_b = z_b


    A().z_b("i am z_b")

    A.load_classmethod("cloudmesh_client.extend.my.my_ccc")
    A().my_ccc("YOURS")
    # A.my_ccc("YOURS")



from .role import AnsibleRole


class AnsibleDeployer(object):

    def __init__(self, roles=None, inventory=None):
        self.roles = roles or []
        self.inventory = inventory

    def generate_playbook(self):
        pass

    def run(self):
        raise NotImplemented



""" run with

python setup.py install; nosetests -v --nocapture tests/cm_basic/test_model.py:Test_model.test_001

nosetests -v --nocapture tests/cm_basic/test_model.py

or

nosetests -v tests/cm_basic/test_model.py

"""
from __future__ import print_function

from pprint import pprint

from cloudmesh_client import CloudmeshDatabase
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.util import HEADING


# noinspection PyMethodMayBeStatic,PyPep8Naming
class Test_model:
    cm = CloudmeshDatabase()

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    # noinspection PyMethodMayBeStatic
    def test_001(self):
        HEADING("cm.tables")
        pprint(self.cm.tables)
        assert True

    def test_002(self):
        HEADING("loop over tablenames")
        for t in self.cm.tables:
            print(t.__tablename__)
        assert "DEFAULT" in str(self.cm.tables)

    def test_003(self):
        HEADING("table info")
        d = self.cm.info()

        print(Printer.write(d))
        assert 'default' in str(d)



# noinspection PyPep8
""" run with

python setup.py install; nosetests -v --nocapture tests/cm_basic/test_database.py:Test_database.test_001

nosetests -v --nocapture tests/cm_basic/test_database.py

or

nosetests -v tests/cm_basic/test_database.py

"""

from pprint import pprint

from cloudmesh_client import CloudmeshDatabase
from cloudmesh_client import DEFAULT
from cloudmesh_client.default import Default
from cloudmesh_client import Printer
from cloudmesh_client.common.util import HEADING


# noinspection PyMethodMayBeStatic,PyPep8Naming
class Test_database:
    cm = CloudmeshDatabase()

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_000(self):
        HEADING("testing DEFAULT add")

        print(Printer.write(self.cm.info()))
        assert True

    def test_001(self):
        HEADING("testing DEFAULT add")

        m = DEFAULT(name="hallo", value="world")
        self.cm.add(m)
        print("added")

        n = self.cm.filter_by(kind="default", name='hallo', scope="first")

        print(n)

        assert n.name == 'hallo'
        assert n.value == 'world'

    def test_002(self):
        HEADING("testing list")

        n = self.cm.find(
            provider='general',
            kind="default",
            scope="first",
            name='hallo')
        pprint(n)

        assert n.name == 'hallo'
        assert n.value == 'world'

    def test_003(self):
        HEADING("testing find")

        m = DEFAULT(name="hallo", value="world")

        n = self.cm.find(kind="default", scope="all", name='hallo')
        print(list(n))
        assert (len(list(n)) > 0)

    def test_004(self):
        print(Printer.write(self.cm.info()))

        m = Default.set_counter("index", 2)
        self.cm.add(m)

        o = Default.get_counter('index')
        print("OOO", o)

        Default.set_counter("index", 0)

        for i in range(0, 10):
            Default.incr_counter("index")

        print(Printer.write(self.cm.info()))

        c = self.cm.all(kind="default")

        print(Printer.write(c, order=['name', 'value', 'provider', 'type']))

        print(Default.get_counter(name="index"))

        i = Default.get(name="index")
        assert type(i) == int
        assert i == 10

        i = Default.index
        assert type(i) == int
        assert i == 10


"""
    def test_005(self):
        

        m = DEFAULT("newfield", "world")
        m.newfield__hhh = 13.9



        self.cm.add(m)


        n = self.cm.find("default", scope="first", name='newfield')

        pprint(n)


    def test_006(self):
        
        m = DEFAULT("other", "world")
        m.other = "ooo"
        self.cm.add(m)

        print("\n\n")
        pprint(self.cm.get(DEFAULT, other='other').__dict__)

"""



from __future__ import print_function

import requests
import hostlist
from builtins import input

from cloudmesh_client.shell.console import Console
from cloudmesh_client.comet.comet import Comet
from cloudmesh_client.common.Printer import Printer

from cloudmesh_client.common.hostlist import Parameter
import time
from datetime import datetime
import pytz
from pprint import pprint


class Cluster(object):
    WALLTIME_MINS = 60 * 24 * 2
    N_ALLOCATIONS_PER_LINE = 5
    MINS_PER_UNIT = {"m": 1, "h": 60, "d": 1440, "w": 10080}
    SECS_PER_DAY = 60 * 60 * 24
    STUCK_COMPUTESETS = ["submitted", "ending"]
    FINISHED_COMPUTESETS = ["completed", "failed"]
    ACTIVE_COMPUTESETS = ["running", "submitted", "created"]
    PENDING_COMPUTESETS = ["queued", "submitted", "created"]
    CLUSTER_ORDER=[
                  "name",
                  "state",
                  "kind",
                  "type",
                  "mac",
                  "ip",
                  "cpus",
                  "cluster",
                  "memory",
                  "disksize",
                  "active_computeset",
                  "allocation",
                  "admin_state"
                    ]
    CLUSTER_HEADER=[
                  "name",
                  "state",
                  "kind",
                  "type",
                  "mac",
                  "ip",
                  "cpus",
                  "cluster",
                  "RAM(M)",
                  "disk(G)",
                  "computeset",
                  "allocation",
                  "admin_state"
                    ]
    '''
    CLUSTER_SORT_KEY=[
                      "name",
                      "state",
                      "kind",
                      "type",
                      "mac",
                      "ip",
                      "cpus",
                      "cluster",
                      "ram",
                      "disk",
                      "computeset",
                      "allocation",
                      "admin_state"]
    '''
    CLUSTER_SORT_KEY=CLUSTER_HEADER[:]
    CLUSTER_SORT_KEY[CLUSTER_HEADER.index("RAM(M)")] = "ram"
    CLUSTER_SORT_KEY[CLUSTER_HEADER.index("disk(G)")] = "disk"

    '''
    NODEINFO_ORDER=[
                    'cpus',
                    'disksize',
                    'dns1',
                    'dns2',
                    'frontend_state',
                    'gateway',
                    'interface',
                    'memory',
                    'name',
                    'ntp',
                    'pub_ip',
                    'pub_mac',
                    'pub_netmask',
                    'state',
                    'type']
    '''

    @staticmethod
    def simple_list(id=None, format="table"):
        result = ""
        if id is None:
            r = Comet.get(Comet.url("cluster/"))
        else:
            r = Comet.get(Comet.url("cluster/" + id + "/"))
            if r is None:
                Console.error("Could not find cluster `{}`"
                              .format(id))
                return result
            r = [r]

        if r is not None:
            if 'error' in r:
                Console.error("An error occurred: {error}".format(**r))
                raise ValueError("COMET Error")
            elif 'error' in r[0]:
                Console.error("An error occurred: {error}".format(**r[0]))
                raise ValueError("COMET Error")

            if format == "rest":
                result = r
            else:
                elements = {}
                for cluster in r:
                    element = {}
                    for attribute in ["project", "name", "description"]:
                        element[attribute] = cluster[attribute]
                        element["nodes"] = len(cluster["computes"])
                    for attribute in cluster["frontend"]:
                        element["frontend " + attribute] = cluster["frontend"][
                            attribute]
                    names = []
                    for compute in cluster["computes"]:
                        names.append(compute["name"])

                    element["computes"] = hostlist.collect_hostlist(names)

                    elements[cluster["name"]] = element

                result = Printer.write(elements,
                                       order=[
                                           "name",
                                           "project",
                                           "nodes",
                                           "computes",
                                           "frontend name",
                                           "frontend state",
                                           "frontend type",
                                           "description",
                                       ],
                                       header=[
                                           "Name",
                                           "Project",
                                           "Count",
                                           "Nodes",
                                           "Frontend (Fe)",
                                           "State (Fe)",
                                           "Type (Fe)",
                                           "Description",
                                       ],

                                       output=format)
            return result

    @staticmethod
    def list(id=None, format="table", sort=None):

        def check_for_error(r):
            if r is not None:
                if 'error' in r:
                    Console.error("An error occurred: {error}".format(**r))
                    raise ValueError("COMET Error")

        result = ""
        if id is None:
            r = Comet.get(Comet.url("cluster/"))
            check_for_error(r)
        else:
            r = Comet.get(Comet.url("cluster/" + id + "/"))
            check_for_error(r)
            if r is None:
                Console.error("Could not find cluster `{}`"
                              .format(id))
                return result
            r = [r]

        #
        # getting account/allocation for each computeset
        # to display in the cluster view
        computeset_account = {}
        # stuck_computesets = {}
        computesets = Comet.get_computeset()
        computesets += Comet.get_computeset(state="submitted")
        computesets += Comet.get_computeset(state="ending")
        # pprint (computesets)
        if computesets:
            for computeset in computesets:
                id = computeset["id"]
                account = computeset["account"]
                if id not in computeset_account:
                    computeset_account[id] = account

        # no longer track and display the (possible) stuck
        # computeset in this way
        '''
        stuck_computesets = {}
        computesets = Comet.get_computeset()
        if computesets:
            for computeset in computesets:
                if computeset["state"] in Cluster.STUCK_COMPUTESETS:
                    cluster = computeset["cluster"]
                    id = computeset["id"]
                    nodes = computeset["computes"]

                    if cluster not in stuck_computesets:
                        stuck_computesets[cluster] = {}
                    for node in nodes:
                        stuck_computesets[cluster][node["name"]] = \
                            "{}({})".format (id, computeset["state"])
        '''

        if r is not None:
            if format == "rest":
                result = r
            else:
                result = ''
                data = []

                empty = {
                    'cluster': None,
                    'cpus': None,
                    'host': None,
                    "mac": None,
                    'ip': None,
                    'memory': None,
                    'disksize': None,
                    'name': None,
                    'state': None,
                    'type': None,
                    'active_computeset': None,
                    'kind': 'frontend',
                    'admin_state': None
                }

                for cluster in sorted(r, key=lambda x: x["name"]):

                    clients = cluster["computes"]
                    for client in clients:
                        client["kind"] = "compute"
                    frontend = dict(empty)
                    frontend.update(cluster["frontend"])
                    pubip = cluster["frontend"]["pub_ip"]
                    frontend["ip"] = pubip
                    frontend["admin_state"] = cluster["frontend"]["frontend_state"]
                    result += "Cluster: %s\tFrontend: %s\tIP: %s\n" % \
                                (cluster["name"],
                                 cluster["frontend"]["name"],
                                 pubip)
                    if len(clients) > 0:
                        frontend['cluster'] = clients[0]['cluster']
                    else:
                        frontend['cluster'] = frontend['name']
                    data += [frontend]
                    data += clients

                for index, anode in enumerate(data):
                    bnode = dict(anode)
                    if "interface" in bnode:
                        macs = []
                        # ips = []
                        for ipaddr in anode["interface"]:
                            macs.append(ipaddr["mac"])
                            # ips.append(ipaddr["ip"] or "N/A")
                        if format == 'table':
                            bnode["mac"] = "\n".join(macs)
                        else:
                            bnode["mac"] = ";".join(macs)

                        if "active_computeset_state" in anode and \
                                anode["active_computeset"] is not None and \
                                anode["active_computeset_state"] is not None:
                            # if not running state, show also the status after
                            # the computeset id
                            if anode["active_computeset_state"] != 'running':
                                bnode["active_computeset"] = "%s(%s)" % \
                                                    (anode["active_computeset"],
                                                     anode["active_computeset_state"])
                            # if an active computeset has a valid account
                            # associated with it, display the account/allocation
                            # does this for computeset in all states
                            if anode["active_computeset"] in computeset_account:
                                bnode["allocation"] = \
                                computeset_account[anode["active_computeset"]]

                        if "compute_state" in anode:
                            bnode["admin_state"] = anode["compute_state"]
                        #anode["ip"] = "; ".join(ips)
                        if "ip" not in bnode:
                            bnode["ip"] = None

                    del bnode["interface"]
                    #
                    # stuck state included in cluster data via API
                    '''
                    if bnode["cluster"] in stuck_computesets and \
                                    bnode["name"] in stuck_computesets[bnode["cluster"]]:
                        bnode["active_computeset"] = \
                            stuck_computesets[bnode["cluster"]][bnode["name"]]
                    '''
                    data[index] = bnode

                sort_keys = ('cluster','mac')
                if sort:
                    if sort in Cluster.CLUSTER_SORT_KEY:
                        # print (sort)
                        idx = Cluster.CLUSTER_SORT_KEY.index(sort)
                        # print (idx)
                        sortkey = Cluster.CLUSTER_ORDER[idx]
                        # print (sortkey)
                        sort_keys = (sortkey,)
                        # print (sort_keys)
                if "table" == format:
                    result_print = Printer.write(data,
                                                 order=Cluster.CLUSTER_ORDER,
                                                 header=Cluster.CLUSTER_HEADER,
                                                 output=format,
                                                 sort_keys=sort_keys)
                    result += str(result_print)
                else:
                    result_print = Printer.write(data,
                                                 order=Cluster.CLUSTER_ORDER,
                                                 header=Cluster.CLUSTER_HEADER,
                                                 output=format,
                                                 sort_keys=sort_keys)
                    result = result_print
            return result

    @staticmethod
    def display_get_allocation(allocations):
        allocation = None
        allocations_sorted = sorted(allocations)
        i = 1
        while i < len(allocations_sorted) + 1:
            for j in range(0, Cluster.N_ALLOCATIONS_PER_LINE):
                if i < len(allocations_sorted) + 1:
                    print ("{}: {}".format(i, allocations_sorted[i - 1]),
                          end="\t")
                    i += 1
            print("")
        print("")
        chosen_alloc = -1
        while chosen_alloc < 0:
            allocation_input = input("Pick an allocation by specifying its index: ")
            try:
                chosen_alloc = int(allocation_input)
                if 0 < chosen_alloc < len(allocations_sorted) + 1:
                    allocation = allocations_sorted[chosen_alloc - 1]
                else:
                    chosen_alloc = -1
                    print("Invalid index specified. "
                          "Please choose between 1 and {}"
                          .format(len(allocations_sorted))
                          )
            except:
                if allocation_input in allocations_sorted:
                    chosen_alloc = 0
                    allocation = allocation_input
        return allocation

    @staticmethod
    def info():
        Console.error("comet cluster info: to be implemented")
        pass

    @staticmethod
    def add():
        pass

    @staticmethod
    def start(id):
        data = {"id": id}
        r = requests.post(Comet.url("cluster/{id}/start".format(**data)))
        print(r)

    @staticmethod
    def stop(id):
        data = {"id": id}
        r = requests.post(Comet.url("cluster/{id}/stop".format(**data)))
        print(r)

    @staticmethod
    def computeset(id=None, cluster=None, state=None, allocation=None):
        #
        # state could be one of
        # ['created' or 'submitted' or 'failed' or 'running' 
        #   or 'cancelled' or 'ending' or 'completed']
        computesets = Comet.get_computeset(id, state)
        if computesets is not None:
            if 'cluster' in computesets:
                result = Cluster.output_computeset(computesets)
            else:
                result = ''
                for acomputeset in computesets:
                    result += Cluster.output_computeset(acomputeset,
                                                        cluster,
                                                        allocation)
        else:
            result = "No computeset exists with the specified ID"
        return result

    @staticmethod
    def output_computeset(computesetdict, cluster=None, allocation=None):
        # print (cluster, state, allocation)
        result = ""
        # filter out based on query criteria
        if cluster and 'cluster' in computesetdict:
            if computesetdict["cluster"] != cluster:
                return result
        # no longer needing state filter as handled by API
        # if state and 'state' in computesetdict:
        #     if "ALL" != state and computesetdict["state"] != state:
        #         return result
        if allocation and 'account' in computesetdict:
            if computesetdict["account"] != allocation:
                return result

        # no longer needing state filter as handled by API
        # if (state and ("ALL" == state or computesetdict["state"] == state)) or \
        #    (computesetdict["state"] not in Cluster.FINISHED_COMPUTESETS):
        starttime = ''
        endtime = ''
        walltime = ''
        runningTime = ''
        remainingTime = ''
        if 'walltime_mins' in computesetdict:
            walltime = computesetdict["walltime_mins"]
            walltime_seconds = walltime * 60
            walltime = Cluster.format_ddd_hh_mm(walltime_seconds)
        if 'start_time' in computesetdict and \
            computesetdict["start_time"] is not None:
            start_seconds = int(computesetdict["start_time"])
            end_seconds = start_seconds + walltime_seconds
            if computesetdict["state"] in Cluster.FINISHED_COMPUTESETS \
                or computesetdict["state"] == 'ending':
                runningSecs = walltime_seconds
            else:
                runningSecs = int(time.time())-start_seconds
            remainingSecs = walltime_seconds - runningSecs
            starttime = time.strftime("%D %H:%M %Z",
                                time.localtime(start_seconds))
            endtime = time.strftime("%D %H:%M %Z",
                                time.localtime(end_seconds))
            runningTime = Cluster.format_ddd_hh_mm(runningSecs)
            remainingTime = Cluster.format_ddd_hh_mm(remainingSecs)

        result += "\nClusterID: {}\tComputesetID: {}\t State: {}\t\tAllocation: {}\n" \
                  "Start (est): {}\t\tEnd (est): {}\n"\
                  "Requested Time (ddd-hh:mm): {}\tRunning Time (est): {}\t\tRemaining Time (est): {}\n"\
                    .format(computesetdict["cluster"],
                            computesetdict["id"],
                            computesetdict["state"],
                            computesetdict["account"],
                            starttime,
                            endtime,
                            walltime,
                            runningTime,
                            remainingTime
                            )
        data = computesetdict["computes"]
        for index, anode in enumerate(data):
            bnode = dict(anode)
            if "interface" in bnode:
                macs = []
                #ips = []
                for ipaddr in anode["interface"]:
                    macs.append(ipaddr["mac"])
                    #ips.append(ipaddr["ip"] or "N/A")
                if format=='table':
                    bnode["mac"] = "\n".join(macs)
                else:
                    bnode["mac"] = ";".join(macs)
                #anode["ip"] = "; ".join(ips)
            del bnode["interface"]
            data[index] = bnode
        result += str(Printer.write(data,
                                   order=[
                                       "name",
                                       "state",
                                       "type",
                                       "mac",
                                       #"ip",
                                       "cpus",
                                       "cluster",
                                       "host",
                                       "memory",
                                   ],
                                   output="table",
                                   sort_keys='cluster'))
        return result

    @staticmethod
    def format_ddd_hh_mm(time_duration_secs):
        ddd = time_duration_secs // Cluster.SECS_PER_DAY
        secs = time_duration_secs % Cluster.SECS_PER_DAY
        hh = secs // 3600
        mm = (secs % 3600) // 60

        if ddd != 0:
            ret = "%s-%02d:%02d" % (ddd, hh, mm)
        else:
            ret = "%02d:%02d" % (hh, mm)
        return ret

    @staticmethod
    def convert_to_mins(s):
        mins = None
        if s is not None:
            s = s.lower()
            if s[-1] in Cluster.MINS_PER_UNIT:
                mins = int(s[:-1]) * Cluster.MINS_PER_UNIT[s[-1]]
        return mins

    @staticmethod
    def check_nodes_computesets(clusterid, computenodeids):
        hosts_param = hostlist.expand_hostlist(computenodeids)
        hosts_param_set = set(hosts_param)
        nodes_free = True
        nodes_allocated = False
        nodes_checked = False
        # computesetid = -1
        computesets = Comet.get_computeset()
        # get all active computeset and put nodes into a set
        allhosts = set()
        for computeset in computesets:
            if computeset["cluster"] == clusterid \
                    and (computeset["state"] in Cluster.ACTIVE_COMPUTESETS):
                computesetid = computeset["id"]
                # print (computesetid)
                for compute in computeset["computes"]:
                    allhosts.add(compute["name"])

        # all nodes allocated
        if hosts_param_set <= allhosts:
            nodes_allocated = True
            nodes_free = False
            nodes_checked = True
        # at least one specified host not in any Active computeset
        else:
            for host in hosts_param:
                # some specified nodes are in Active computeset
                if host in allhosts:
                    nodes_free = False
                    nodes_checked = True
                    break

        # print ("nodes_checked: %s" % nodes_checked)
        # print ("nodes_free: %s" % nodes_free)
        return [nodes_free, nodes_allocated]

    @staticmethod
    def computeset_start(clusterid,
                         computenodeids=None,
                         numnodes=None,
                         allocation=None,
                         walltime=None):
        ret = ''
        # print ("clusterid: %s" % clusterid)
        # print ("computenodeids: %s" % computenodeids)
        # print ("numnodes: %s" % numnodes)
        # print ("allocation: %s" % allocation)
        # print ("walltime: %s" % walltime)
        data = None
        # validating and initiating allocation and walltime values
        if not allocation:
            cluster = Cluster.list(clusterid, format='rest')
            # use the first one if no provided
            allocation = cluster[0]['allocations'][0]
        if not walltime:
            walltime = Cluster.WALLTIME_MINS

        # preparing parameters for the post call
        # num of nodes to start
        #
        # Now it accepts
        # {"cluster":"vc3","computes":"compute[1-2]"},
        # {"cluster":"vc3","computes":["compute1","compute2"]} and
        # {"cluster":"vc3","count":2}
        #
        if numnodes:
            data = {"cluster": "%s" % clusterid,
                    "count": "%s" % numnodes,
                    "walltime_mins": "%s" % walltime,
                    "allocation": "%s" % allocation}
        # computenodeids in hostlist format
        elif computenodeids:
            nodes_free = Cluster.check_nodes_computesets(clusterid, computenodeids)[0]
            if not nodes_free:
                Console.error("Some nodes are already in active computesets",
                              traceflag=False)
                return ret
            else:
                data = {"computes": "%s" % computenodeids,
                        "cluster": "%s" % clusterid,
                        "walltime_mins": "%s" % walltime,
                        "allocation": "%s" % allocation}

        # print("Issuing request to poweron nodes...")
        posturl = Comet.url("computeset/")
        # print ("POST data: %s" % data)
        r = Comet.post(posturl, data=data)
        # print("RETURNED RESULTS:")
        # print (r)
        if r is None:
            ret = "An error occured. Please check your command and try again"
        else:
            if 'cluster' in r:
                if 'state' in r and \
                                r['state'] in Cluster.PENDING_COMPUTESETS:
                    computesetid = r['id']
                    ret = 'Request accepted! Check status with:\n' \
                          'comet cluster {}\n'.format(clusterid) + \
                          'or:\n' \
                          'comet computeset {}\n'.format(computesetid)
                else:
                    # in case of some internal problem
                    ret = ''
            elif 'error' in r:
                ret = "An error occurred: {}".format(r['error'])
            else:
                ret = "An internal error occured. " \
                      "Please submit a ticket with the " \
                      "following info:\n {}\n" \
                    .format(r)
        return ret

    @staticmethod
    def computeset_terminate(computesetid):
        ret = ''
        url = Comet.url("computeset/")
        action = "shutdown"
        puturl = "{:}{:}/{}".format(url, computesetid, action)
        # print (puturl)
        r = Comet.put(puturl)
        if r is not None:
            if '' != r.strip():
                ret = r
            else:
                ret = "Request Accepted. " \
                      "In the process of terminating the computeset"
        else:
            ret = "Problem executing the request. " \
                  "Check if the computeset exists"
        return ret

    @staticmethod
    def power(clusterid, subject, param=None, action=None):
        ret = ''
        # print ("clusterid: %s" % clusterid)
        # print ("subject: %s" % subject)
        # print ("param: %s" % param)
        # print ("action: %s" % action)
        if subject in ['HOSTS']:
            nodes_allocated = Cluster.check_nodes_computesets(clusterid, param)[1]
            if nodes_allocated:
                hosts_param = hostlist.expand_hostlist(param)
                # print (hosts_param)
                for host in hosts_param:
                    url = Comet.url("cluster/{}/compute/{}"
                                    .format(clusterid, host))
                    if action in ["on", "off"]:
                        action = "power{}".format(action)
                    puturl = "{}/{}".format(url, action)
                    # print (puturl)
                    r = Comet.put(puturl)
                    if r is not None:
                        if '' != r.strip():
                            ret = r
                        else:
                            ret += "Request Accepted. " \
                                   "In the process of {} node {}\n" \
                                .format(action, host)
                    else:
                        ret += "Problem executing the request. " \
                               "Check if the node {} belongs to the cluster" \
                            .format(host)
                        # print(ret)
            else:
                Console.error("At least one specified node is not in any "
                              "active computesets thus cannot be powered on.\n"
                              "Please start the node(s) with 'comet start' first",
                              traceflag=False)
        elif 'FE' == subject:
            url = Comet.url("cluster/{}/frontend/".format(clusterid))
            if action in ["on", "off", "reboot", "reset", "shutdown"]:
                if action in ["on", "off"]:
                    action = "power{}".format(action)
                puturl = "{}{}".format(url, action)
                # print (puturl)
                r = Comet.put(puturl)
                if r is not None:
                    if '' != r.strip():
                        ret = r
                    else:
                        ret = "Request Accepted. " \
                              "In the process of {} the front-end node".format(action)
                else:
                    ret = "Problem executing the request. " \
                          "Check if the cluster exists"
            else:
                ret = "Action not supported! Try these: on/off/reboot/reset/shutdown"
        elif 'COMPUTESET' == subject:
            url = Comet.url("computeset/")
            if action in ["on", "off"]:
                ret = "Action NOT SUPPORTED! Use 'comet start/terminate'!"
            elif action in ["reboot", "reset", "shutdown"]:
                if action in ["off"]:
                    action = "power{}".format(action)
                puturl = "{:}{:}/{}".format(url, param, action)
                # print (puturl)
                r = Comet.put(puturl)
                if r is not None:
                    if '' != r.strip():
                        ret = r
                    else:
                        ret = "Request Accepted. " \
                              "In the process of {} the computeset".format(action)
                else:
                    ret = "Problem executing the request. " \
                          "Check if the computeset exists"
            else:
                ret = "Action not supported! Try these: reboot/reset/shutdown"
        return ret

    @staticmethod
    def detach_iso(clusterid, nodeids=None, action='Detaching'):
        return Cluster.attach_iso('', clusterid, nodeids, action='Detaching')

    @staticmethod
    def attach_iso(iso_id_name, clusterid, nodeids=None, action='Attaching'):
        ret = ''
        # print ("Attaching ISO image")
        # print ("iso_id_name: %s" % iso_id_name)
        # print ("cluster: %s" % clusterid)
        # print ("node: %s" % nodeid)

        # obtain the list of iso first to support attach by index
        # and easy verification of image name
        isodict = {}
        isos = (Comet.list_iso())
        idx = 0
        for iso in isos:
            if iso.startswith("public/"):
                iso = iso.split("/")[1]
            idx += 1
            isodict[str(idx)] = iso

        if iso_id_name != '':
            if iso_id_name.isdigit():
                if iso_id_name in isodict:
                    iso_id_name = isodict[iso_id_name]
                else:
                    iso_id_name = None
            else:
                if not iso_id_name in list(isodict.values()):
                    iso_id_name = None

        if iso_id_name is not None:
            # currently only those in public directory is accessible
            # for empty string, DO NOT append as that is for detach
            if len(iso_id_name) != 0:
                iso_id_name = "public/{}".format(iso_id_name)
            urls = {}
            # attaching to compute node
            if nodeids:
                nodeids = hostlist.expand_hostlist(nodeids)
                for nodeid in nodeids:
                    url = Comet.url("cluster/{}/compute/{}/attach_iso?iso_name={}") \
                        .format(clusterid, nodeid, iso_id_name)
                    urls["Node {}".format(nodeid)] = url
            else:
                # attaching to fronend node
                url = Comet.url("cluster/{}/frontend/attach_iso?iso_name={}") \
                    .format(clusterid, iso_id_name)
                urls['Frontend'] = url
            # data = {"iso_name": "%s" % iso_id_name}
            # print ("url: %s" % url)
            # print ("data: %s" % data)
            tofrom = {}
            tofrom['Attaching'] = 'to'
            tofrom['Detaching'] = 'from'
            for node, url in urls.items():
                url = url.replace("+","%2B")
                r = Comet.put(url)
                # print (r)
                if r is not None:
                    if '' != r.strip():
                        ret += r
                    else:
                        ret += "Request Accepted. {} the iso image {} {} of cluster {}\n" \
                            .format(action, tofrom[action], node, clusterid)
                else:
                    ret += "Something wrong during {} the iso image {} {} of cluster {}!" \
                           "Please check the command and try again\n" \
                        .format(action, tofrom[action], node, clusterid)
        else:
            ret = "ERROR: The specified index or image name not matching any from the ISO list!\n"\
                  "Check 'cm comet iso list'"
        return ret

    @staticmethod
    def rename_node(clusterid, old_compute_name, new_compute_name):
        url = Comet.url("cluster/{}/compute/{}/rename"
                        .format(clusterid, old_compute_name))
        data = {"name": "%s" % new_compute_name}
        ret = ""
        r = Comet.post(url, data=data)
        # print (r)
        if r is not None:
            if '' != r.strip():
                ret = r
            else:
                ret = "Request Accepted."
        return ret

    @staticmethod
    def node_info(clusterid, nodeid=None, format='table'):
        if nodeid:
            url = Comet.url("cluster/{}/compute/{}/"
                            .format(clusterid, nodeid))
        else:
            url = Comet.url("cluster/{}/frontend/"
                            .format(clusterid))
        r = Comet.get(url)
        return Printer.attribute(r,
                                 #order=Cluster.NODEINFO_ORDER,
                                 output=format)

    @staticmethod
    def delete():
        pass



from __future__ import print_function

from cloudmesh_client.shell.console import Console
from cloudmesh_client.shell.command import command, PluginCommand, \
    CloudPluginCommand, CometPluginCommand
from cloudmesh_client.default import Default
from cloudmesh_client.common.LogUtil import LogUtil

logger = LogUtil.get_logger()


class DefaultCommand(PluginCommand, CloudPluginCommand, CometPluginCommand):
    topics = {"default": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command default")

    # noinspection PyUnusedLocal
    @command
    def do_default(self, args, arguments):
        """
        ::

          Usage:
              default
              default list [--cloud=CLOUD] [--format=FORMAT] [--all]
              default delete KEY [--cloud=CLOUD]
              default KEY [--cloud=CLOUD]
              default KEY=VALUE [--cloud=CLOUD]

          Arguments:
            KEY    the name of the default
            VALUE  the value to set the key to

          Options:
             --cloud=CLOUD    the name of the cloud
             --format=FORMAT  the output format. Values include
                              table, json, csv, yaml.
             --all            lists all the default values

        Description:
            Cloudmesh has the ability to manage easily multiple
            clouds. One of the key concepts to manage multiple clouds
            is to use defaults for the cloud, the images, flavors,
            and other values. The default command is used to manage
            such default values. These defaults are used in other commands
            if they are not overwritten by a command parameter.

            The current default values can by listed with

                default list --all

            Via the default command you can list, set, get and delete
            default values. You can list the defaults with

               default list

            A default can be set with

                default KEY=VALUE

            To look up a default value you can say

                default KEY

            A default can be deleted with

                default delete KEY

            To be specific to a cloud you can specify the name of the
            cloud with the --cloud=CLOUD option. The list command can
            print the information in various formats iv specified.

        Examples:
            default
                lists the default for the current default cloud

            default list --all
                lists all default values

            default list --cloud=kilo
                lists the defaults for the cloud with the name kilo

            default image=xyz
                sets the default image for the default cloud to xyz

            default image=abc --cloud=kilo
                sets the default image for the cloud kilo to xyz

            default image
                list the default image of the default cloud

            default image --cloud=kilo
                list the default image of the cloud kilo

            default delete image
                deletes the value for the default image in the
                default cloud

            default delete image --cloud=kilo
                deletes the value for the default image in the
                cloud kilo

        """
        #print(arguments)
        #print (">", args, "<")

        """
        For these keys, the 'cloud' column in db
        will always be 'general'.
        """
        general_keys = ["cloud", "cluster", "queue", "key", "group", "user", "secgroup", "vm",
                        "refresh", "debug", "interactive", "purge"]

        """
        If the default cloud has been set (eg. default category=xxx),
        then subsequent defaults for any key (eg. default image=yyy),
        will have 'cloud' column in db as the default cloud that was set.
        (eg. image=yyy for category=xxx).
        """

        if arguments["KEY"] in general_keys:
            cloud = "general"
        elif args == '':
            cloud = "general"
            arguments["--cloud"] = cloud
            arguments["list"] = True
            order = ['name', 'value']
            output_format = arguments["--format"]
            result = Default.list(category=cloud,
                                  order=order,
                                  output=output_format)
            print(result)
            return ""

        else:
            cloud = arguments["--cloud"] or Default.get(name="cloud", category="general") or "general"

        if arguments["list"]:
            output_format = arguments["--format"] or Default.output or 'table'

            if arguments['--all'] or arguments["--cloud"] is None:
                cloud = None
            result = Default.list(category=cloud, output=output_format)

            if result is None:
                Console.error("No default values found")
            else:
                print(result)
            return ""

        elif arguments["delete"]:

            key = arguments["KEY"]
            if key in general_keys:
                cloud = "general"
            result = Default.delete(key, cloud)
            if not result :
                Console.error("default {} not present".format(key))
            else:
                Console.ok("Deleted key {} for cloud {}. ok.".format(key,
                                                                     cloud))
            return ""

        elif "=" in arguments["KEY"]:
            key, value = arguments["KEY"].split("=")
            if key in general_keys:
                cloud = "general"
            if key in "debug":
                Default.set_debug(value)
            else:
                Default.set(key, value, category=cloud)
            Console.ok(
                "set default {}={}. ok.".format(key, value))
            return ""

        elif arguments["KEY"]:
            key = arguments["KEY"]
            if key in general_keys:
                cloud = "general"
            result = Default.get(name=key, category=cloud)
            if result is None:
                Console.error("No default values found")
            else:
                Console.ok("{}".format(result))
            return ""



from .general.attributes import Attributes as GeneralAttributes
from .openstack.attributes import Attributes as OpenstackAttributes
from .libcloud.attributes import Attributes as LibcloudAttributes


class Attributes(object):
    @classmethod
    def get(cls, kind, provider='general'):
        if provider == "general":
            return GeneralAttributes.get(kind)
        elif provider == "openstack":
            return OpenstackAttributes.get(kind)
        elif provider == "libcloud":
            return LibcloudAttributes.get(kind)
        else:
            ValueError("provider {} not found".format(provider))
            return None, None



from __future__ import print_function
from cloudmesh_client.cloud.workflow import Workflow
from cloudmesh_client.shell.command import command, PluginCommand, CloudPluginCommand
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default
from pprint import pprint
from cloudmesh_client.common.hostlist import Parameter
from cloudmesh_client.common.dotdict import dotdict
# from cloudmesh_client.cloud.workflowexec import *

class WorkflowCommand(PluginCommand, CloudPluginCommand):
    topics = {"workflow": "todo"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command workflow")

    # noinspection PyUnusedLocal
    @command
    def do_workflow(self, args, arguments):
        """
        ::

            Usage:
                workflow refresh [--cloud=CLOUD] [-v]
                workflow list [ID] [NAME] [--cloud=CLOUD] [--format=FORMAT] [--refresh] [-v]
                workflow add NAME LOCATION
                workflow delete ID
                workflow status [NAMES]
                workflow show ID
                workflow save NAME WORKFLOWSTR
                workflow run NAME
                workflow service start
                workflow service stop
                This lists out the workflows present for a cloud

            Options:
               --format=FORMAT  the output format [default: table]
               --cloud=CLOUD    the cloud name
               --refresh        refreshes the data before displaying it
                                from the cloud

            Examples:
                cm workflow refresh
                cm workflow list
                cm workflow list --format=csv
                cm workflow show 58c9552c-8d93-42c0-9dea-5f48d90a3188 --refresh
                cm workflow run workflow1

        """


        arg = dotdict(arguments)
        if arg.NAMES is not None:
            arg.names = Parameter.expand(arg.NAMES)
        else:
            arg.names = None
        pprint (arg)



        cloud = arguments["--cloud"] or Default.cloud
        if cloud is None:
            Console.error("Default cloud doesn't exist")
            return

        if arguments["-v"]:
            print("Cloud: {}".format(cloud))

        if arguments["refresh"] or Default.refresh:
            msg = "Refresh workflow for cloud {:}.".format(cloud)
            if Workflow.refresh(cloud):
                Console.ok("{:} ok".format(msg))
            else:
                Console.error("{:} failed".format(msg))
                return ""

        if arguments["list"]:
            id = arguments['ID']
            live = arguments['--refresh']
            output_format = arguments["--format"]

            result = None
            if id is None:
                result = Workflow.list(cloud, output_format)
            else:
                result = Workflow.details(cloud, id, live, output_format)

            if result is None:
                Console.error("No workflow(s) found. Failed.")
            else:
                print(result)
            return ""

        elif arguments["show"]:
            workflow_id = arguments["ID"]
            Console.ok("I executed show")
            # Console.msg(workflow_id)
            if workflow_id is None:
                Console.msg("Please enter a Workflow Id to execute workflow")
            else:
                result = Workflow.run(cloud,workflow_id)
                # print (result)
                if result is None:
                    Console.msg("Use workflow list to view existing workflows or do workflow save to save a new one")
                else:
                    # Console.msg(result)
                    # Console.msg(result[0])
                    Console.msg(result[0]['workflow_str'])
                    # entry_point(result[0]['workflow_str'])
                    Console.msg("All Set to execute")

        elif arguments["save"]:
            workflow_name = arguments["NAME"]
            workflow_str = arguments["WORKFLOWSTR"]

            result = Workflow.save(cloud=cloud,
                                   name=workflow_name,
                                   str=workflow_str)
            if result is not None:
                Console.ok(result)
            else:
                Console.error("Failed to save workflow!")
        elif arguments["run"]:
            Console.msg("Execute Run")
            workflow_name = arguments["NAME"]
            # cm_id = arguments["ID"]

            result = Workflow.run(cloud,name=workflow_name)

        elif arguments["delete"]:
            workflow_id = arguments["ID"]

            result = Workflow.delete(cloud,workflow_id)
            if result is not None:
                Console.msg(result)
            else:
                Console.error("Failed to save workflow!")
        elif arguments["service"] and arguments["start"]:
            # Console.msg()
            # mkdir -pf ~/.cloudmesh/workflow
            pass


""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_clouds/test_chameleon.py:Test_chameleon.test_001

nosetests -v --nocapture tests/test_image.py

or

nosetests -v tests/test_chameleon.py

"""

import os

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner


# noinspection PyMethodMayBeStatic,PyPep8Naming
class Test_chameleon:
    """
        This class tests the ImageCommand
    """

    # noinspection PyTypeChecker
    data = dotdict({
        "cloud": "chameleon",
    })

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return result

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_001(self):
        """
        test image refresh
        :return:
        """
        HEADING()
        result = self.run("make db")
        assert "ok." in result

    def test_002(self):
        """
        test image refresh
        :return:
        """
        HEADING()
        result = self.run("cm default cloud={cloud}")
        assert "ok." in result

    def test_003(self):
        """
        test image refresh
        :return:
        """
        HEADING()
        os.system("py.test tests/cm_cloud")
        # assert "ok." in result
        assert True



from __future__ import print_function
from cloudmesh_client.db.CloudmeshDatabase import CloudmeshDatabase, CloudmeshMixin

from sqlalchemy import Column, Date, Integer, String, Boolean
from sqlalchemy.orm.session import object_session


class WORKFLOW(CloudmeshMixin, CloudmeshDatabase.Base):
    """table to store default values

    if the category is "global" it is meant to be a global variable

    todo: check if its global or general
    """
    __tablename__ = "workflow"
    __kind__ = 'workflow'
    __provider__ = 'general'

    status = Column(String)
    location = Column(String)
    description = Column(String)
    dependency = Column(String)
    threads = Column(String) # we store this as string, you need to convert to and from string to use

    def __init__(self, **kwargs):
        super(WORKFLOW, self).set_defaults(**kwargs)
        self.status = kwargs.get('status', 'defined')
        self.description = kwargs.get('description', None)
        self.dependency = kwargs.get('dependency', None)
        self.threads = kwargs.get('threads', '1')


class DEFAULT(CloudmeshMixin, CloudmeshDatabase.Base):
    """table to store default values

    if the category is "global" it is meant to be a global variable

    todo: check if its global or general
    """
    __tablename__ = "default"
    __kind__ = 'default'
    __provider__ = 'general'

    value = Column(String)

    def __init__(self, **kwargs):
        super(DEFAULT, self).set_defaults(**kwargs)
        self.value = kwargs.get('value', None)


class VAR(CloudmeshMixin, CloudmeshDatabase.Base):
    """table to store peristant variable values
    """
    # name defined in mixin

    __tablename__ = "var"

    __kind__ = 'var'
    __provider__ = 'general'

    value = Column(String)

    def __init__(self, **kwargs):
        super(VAR, self).set_defaults(**kwargs)

        self.value = str(kwargs.get('value'))


class LAUNCHER(CloudmeshMixin, CloudmeshDatabase.Base):
    """table to store default values

    if the category is "global" it is meant to be a global variable

    todo: check if its global or general
    """

    __tablename__ = "launcher"

    __kind__ = 'launcher'
    __provider__ = 'general'

    parameters = Column(String)  # This is the parameter represented as yaml object
    source = Column(String)

    def __init__(self, **kwargs):
        super(LAUNCHER, self).set_defaults(**kwargs)
        self.parameters = kwargs.get('parameters')
        self.source = kwargs.get('source')


class KEY(CloudmeshMixin, CloudmeshDatabase.Base):
    __tablename__ = "key"

    __kind__ = 'key'
    __provider__ = 'general'

    fingerprint = Column(String, unique=True)
    value = Column(String)
    source = Column(String)
    comment = Column(String)
    uri = Column(String)
    is_default = Column(String)

    def __init__(self, **kwargs):
        super(KEY, self).set_defaults(**kwargs)

        self.value = kwargs.get("value")
        self.uri = kwargs.get("uri")
        self.comment = kwargs.get("comment")
        self.fingerprint = kwargs.get("fingerprint")
        self.source = kwargs.get("source")
        self.is_default = kwargs.get("is_default")


class GROUP(CloudmeshMixin, CloudmeshDatabase.Base):
    __tablename__ = "group"

    __kind__ = 'group'
    __provider__ = 'general'

    member = Column(String)
    species = Column(String)

    def __init__(self, **kwargs):
        super(GROUP, self).set_defaults(**kwargs)

        self.member = kwargs.get('member')
        self.species = kwargs.get('species') or 'vm'
        self.category = kwargs.get('category', 'general')
        self.type = 'str'


class RESERVATION(CloudmeshMixin, CloudmeshDatabase.Base):
    __tablename__ = "reservation"

    __kind__ = 'reservation'
    __provider__ = 'general'

    hosts = Column(String)  # should be list of strings
    description = Column(String)
    start_time = Column(String)  # date, time
    end_time = Column(String)  # date, time

    def __init__(self, **kwargs):
        super(RESERVATION, self).set_defaults(**kwargs)
        self.hosts = kwargs.get("hosts")
        self.start_time = kwargs.get("start")
        self.end_time = kwargs.get("end")
        self.description = kwargs.get("description")
        self.project = kwargs.get("project")


class SECGROUPRULE(CloudmeshMixin, CloudmeshDatabase.Base):
    __tablename__ = "secgrouprule"

    __kind__ = 'secgrouprule'
    __provider__ = 'general'

    group = Column(String)  # What is this is this name?
    fromPort = Column(String)
    toPort = Column(String)
    protocol = Column(String)
    cidr = Column(String)
    uuid = Column(String)

    # noinspection PyPep8Naming
    def __init__(self, **kwargs):
        super(SECGROUPRULE, self).set_defaults(**kwargs)
        self.uuid = kwargs.get("uuid")
        self.group = kwargs.get("group")  # What is this is this name?
        self.project = kwargs.get("project")
        self.fromPort = kwargs.get("fromPort")
        self.toPort = kwargs.get("toPort")
        self.protocol = kwargs.get("protocol")
        self.cidr = kwargs.get("cidr")


class BATCHJOB(CloudmeshMixin, CloudmeshDatabase.Base):
    __tablename__ = "batchjob"

    __kind__ = 'batchjob'
    __provider__ = 'general'

    dir = Column(String, default="string")
    nodes = Column(String, default="string")
    output_file = Column(String, default="string")
    queue = Column(String, default="string")
    time = Column(String, default="string")
    cluster = Column(String, default="string")
    sbatch_file_path = Column(String, default="string")
    cmd = Column(String, default="string")
    time = Column(String, default="string")
    group = Column(String, default="string")
    job_id = Column(String, default="string")

    def __init__(self, **kwargs):
        super(BATCHJOB, self).set_defaults(name=name, user=user)
        self.dir = kwargs.get('dir')
        self.nodes = kwargs.get('nodes')
        self.output_file = kwargs.pop('output_file')
        self.queue = kwargs.pop('queue')
        self.time = kwargs.pop('time')
        self.cluster = kwargs.pop('cluster')
        self.sbatch_file_path = kwargs.pop('sbatch_file_path')
        self.cmd = kwargs.pop('cmd')
        self.time = kwargs.pop('time')
        self.group = kwargs.pop('group')
        self.job_id = kwargs.pop('job_id')


class CLUSTER(CloudmeshMixin, CloudmeshDatabase.Base):
    """This is a proxy for the main implementation, given in
    :module:`cloudmesh_client.cloud.cluster`.

    The implementation is split from the definitions in order to avoid
    recursive dependencies. If the implementation of various methods
    were done withing this class, various helper functions would need
    to be imported (eg :module:`cloudmesh_client.cloud.vm`). These
    module import :module:`cloudmesh_client.db.CloudmeshDatabase`,
    which import the various ``models`` modules. When this happens,
    the error that is an ImportError for CloudmeshDatabase.

    When modifying, add attributes to this class, then update the
    implementation in subclass.
    """

    __tablename__ = 'cluster'
    __kind__ = 'cluster'
    __provider__ = 'general'

    name = Column(String, unique=True)
    count = Column(Integer, default=1)
    cloud = Column(String)
    username = Column(String)
    image = Column(String)
    flavor = Column(String)
    key = Column(String)
    secgroupname = Column(String)
    assignFloatingIP = Column(Boolean, default=True)

    def __init__(self, name=None, cloud=None, count=None,
                 username=None, image=None, flavor=None,
                 key=None, secgroup=None,
                 assignFloatingIP=True):

        assert cloud is not None
        assert image is not None
        assert flavor is not None
        assert key is not None

        self.name = name
        self.count = count
        self.cloud = cloud
        self.username = username
        self.image = image
        self.flavor = flavor
        self.key = key
        self.secgroup = secgroup
        self.assignFloatingIP = assignFloatingIP




from __future__ import print_function
import subprocess

from cloudmesh_client.shell.console import Console
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.db import CloudmeshDatabase
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.cloud.ListResource import ListResource

# change test gergor
# noinspection PyUnusedLocal
class Launcher(ListResource):

    cm = CloudmeshDatabase()

    def info(self, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def list(cls, name=None, output='table'):
        # ignore names for now
        try:
            elements = cls.cm.find(kind="launcher", category='general', scope="all", output="dict")
            order = None
            header = None

            return Printer.write(elements,
                                 order=order,
                                 header=header,
                                 output=output)
        except Exception as ex:
            Console.error(ex.message)
            return ""

    @classmethod
    def add(cls, name=None, source=None):
        d = {
            "kind": "launcher",
            "provider": "general",
            "category": "general",
            "name": name,
            "source": source,
            "parameter": "We find that in source"
        }

        cls.cm.add(d)
        return str(d.name)

    @classmethod
    def delete(cls, name=None, category=None):
        try:
            cls.cm.delete(kind="launcher", provider="general", name=name)
        except Exception as ex:
            Console.error(ex.message)
            return ""

    @classmethod
    def run(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def resume(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def suspend(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def kill(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def details(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def clear(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

    @classmethod
    def refresh(cls, **kwargs):
        Console.TBD("not yet implemented")
        return "not yet implemented"

"""
class LauncherShell(LauncherProvider):
    def __init__(self):
        pass

    def run(self, **kwargs):
        script = kwargs["script"].strip("\n")

        print(">>>>", script)
        # output = Shell.sh(script)
        output = subprocess.check_output(script, shell=True)
        print("OOOOOO", output)

        return output
"""

"""
# noinspection PyPep8Naming
def Launcher(kind):


    #if kind.lower() in ["sh", "shell"]:
    #    return LauncherShell()
    #else:
        Console.TBD("not yet implemented")
        return "not yet implemented"
"""


from __future__ import print_function

import socket
from uuid import UUID

from cloudmesh_client.shell.console import Console
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.cloud.ListResource import ListResource
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.db.CloudmeshDatabase import CloudmeshDatabase

from pprint import pprint
from builtins import input


# noinspection PyBroadException,PyPep8Naming
class Ip(ListResource):
    cm = CloudmeshDatabase()

    @classmethod
    def list(cls, cloud=None, names=None, output='table', live=False):

        try:

            if live:
                cls.refresh(cloud)

            elements = cls.cm.find(kind="vm", category=cloud)
            result = []
            if "all" in names:
                for element in elements:
                    result.append(element)
            elif names is not None:
                for element in elements:
                    if element["name"] in names:
                        result.append(element)

            (order, header) = CloudProvider(cloud).get_attributes("ip")

            return Printer.write(result,
                                 order=order,
                                 header=header,
                                 output=output)
        except Exception as ex:
            Console.error(ex.message)





            # def list_floating_ip_pools(self, **kwargs):
            #        return self._to_dict(self.provider.floating_ip_pools.list())



""" run with

python setup.py install; nosetests -v --nocapture tests/cm_basic/test_var.py
python setup.py install; nosetests -v --nocapture tests/cm_basic/test_var.py:Test_var.test_001

nosetests -v --nocapture tests/cm_basic/test_var.py

or

nosetests -v tests/cm_basic/test_var.py

"""
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.util import HEADING

from cloudmesh_client.var import Var


def run(command):
    print(command)
    parameter = command.split(" ")
    shell_command = parameter[0]
    args = parameter[1:]
    result = Shell.execute(shell_command, args)
    print(result)
    return result


# noinspection PyMethodMayBeStatic,PyPep8Naming
class Test_var(object):
    """

    """

    def setup(self):
        pass

    def test_001(self):
        HEADING(" delete defaults")
        Var.clear()
        print(Var.list())
        assert Var.list() is None

    def _check(self, content):
        result = Var.list()
        print(result)
        assert content in str(result)

    def test_002(self):
        HEADING("set default variable")
        name = "myvar"
        value = "myvalue"
        Var.set(name, value)
        print(Var.list())
        self._check(value)

    def test_003(self):
        HEADING("delete default variable")
        Var.clear()
        name = "deleteme"
        value = "myvalue"
        Var.set(name, value)
        print(Var.list())
        Var.delete(name)
        print(Var.list())
        self._check("None")

    def test_004(self):
        HEADING("assign a variable")
        command = "cm var myvar=myvalue"
        result = run(command)
        result = run("cm var list")
        assert "myvalue" in result

    def test_005(self):
        HEADING("print a variable")
        command = "cm banner $myvar"
        result = run(command)
        assert "myvalue" in result
        assert "myvar" in result


    def test_999(self):
        HEADING("clear the defaults")

        Var.clear()
        result = Var.list()
        print(result)
        assert result is None

    '''
    def test_002(self):
        HEADING("tries to start a vm with an invalid image")

        result = run ("cm vm start --cloud=india --flavor=m1.medium --image=futuresystems/linux>windows")

        assert "not found" in result
    '''



from __future__ import print_function
import json

import yaml

from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.common.ConfigDict import Config
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.default import Default
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.common.Error import Error
from cloudmesh_client.cloud.key import Key

from cloudmesh_client.common.dotdict import dotdict
from pprint import pprint


class KeyCommand(PluginCommand, CloudPluginCommand):
    topics = {"key": "security"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command key")

    # noinspection PyUnusedLocal
    @command
    def do_key(self, args, arguments):
        """
        ::

           Usage:
             key  -h | --help
             key list --cloud=CLOUD
             key list --source=db [--format=FORMAT]
             key list --source=yaml [--format=FORMAT]
             key list --source=ssh [--dir=DIR] [--format=FORMAT]
             key list --source=git [--format=FORMAT] [--username=USERNAME]
             key list
             key load [--format=FORMAT]
             key add [NAME] [--source=FILENAME]
             key add [NAME] [--git]
             key add [NAME] [--ssh]
             key get NAME
             key default --select
             key delete (NAME | --select | --all)
             key delete NAME --cloud=CLOUD
             key upload [NAME] [--cloud=CLOUD]
             key upload [NAME] --active

           Manages the keys

           Arguments:

             CLOUD          The cloud
             NAME           The name of the key.
             SOURCE         db, ssh, all
             KEYNAME        The name of a key. For key upload it defaults to the default key name.
             FORMAT         The format of the output (table, json, yaml)
             FILENAME       The filename with full path in which the key
                            is located
             NAME_ON_CLOUD  Typically the name of the keypair on the cloud.

           Options:

              --dir=DIR                     the directory with keys [default: ~/.ssh]
              --format=FORMAT               the format of the output [default: table]
              --source=SOURCE               the source for the keys [default: db]
              --username=USERNAME           the source for the keys [default: none]
              --name=KEYNAME                The name of a key
              --all                         delete all keys
              --force                       delete the key form the cloud
              --name_on_cloud=NAME_ON_CLOUD Typically the name of the keypair on the cloud.

           Description:

           key list --source=git  [--username=USERNAME]

              lists all keys in git for the specified user. If the
              name is not specified it is read from cloudmesh.yaml

           key list --source=ssh  [--dir=DIR] [--format=FORMAT]

              lists all keys in the directory. If the directory is not
              specified the default will be ~/.ssh

           key list --source=yaml  [--dir=DIR] [--format=FORMAT]

              lists all keys in cloudmesh.yaml file in the specified directory.
               dir is by default ~/.cloudmesh

           key list [--format=FORMAT]

               list the keys in the giiven format: json, yaml,
               table. table is default

           key list

                Prints list of keys. NAME of the key can be specified

           key add ssh

               adds the default key with the name id_rsa.pub

           key add NAME  --source=FILENAME

               adds the key specifid by the filename to the key
               database

           key get NAME

               Retrieves the key indicated by the NAME parameter from database
               and prints its fingerprint.

           key default --select

                Select the default key interactively

           key delete NAME

                deletes a key. In yaml mode it can delete only key that
                are not saved in the database

           key rename NAME NEW

                renames the key from NAME to NEW.
                
        """
        # pprint(arguments)

        invalid_names = ['tbd', 'none', "", 'id_rsa']

        def _print_dict(d, header=None, format='table'):
            msg = Printer.write(d,
                                 order=["name",
                                        "comment",
                                        "uri",
                                        "fingerprint",
                                        "source"],
                                 output=format,
                                 sort_keys=True)
            if msg is None:
                Console.error("No keys found.", traceflag=False)
                return None
            else:
                return msg

        directory = Config.path_expand(arguments["--dir"])

        cloud = arguments["--cloud"] or Default.cloud

        if arguments['list']:
            _format = arguments['--format']
            _source = arguments['--source']
            _dir = arguments['--dir']

            if "--source" not in arguments and "--cloud" not in arguments:
                arguments["--source"] = 'db'


            if arguments['--cloud']:

                    #
                    # get key list from openstack cloud
                    #
                    #keys = Key.list(cloud, output=_format)

                    #keys = Key.list_on_cloud(cloud, live=True, format=_format)
                    keys = Key.get_from_cloud(cloud, live=True, format=_format)
                    if keys is None:
                        Console.ok("The Key list is empty")
                    else:
                        print(Printer.write(keys,
                            order=["name",
                                "fingerprint"],
                            output=_format or "table"))
                    return ""

            elif arguments['--source'] == 'ssh':

                try:
                    #sshm = SSHKeyManager()
                    d = Key.get_from_dir(directory, store=False)

                    #print("SSS", type(Key.__keys__))
                    #d = dict(Key.all())
                    #print(d)
                    print(Printer.write(d,
                                        order=["name",
                                               "comment",
                                               "uri",
                                               "fingerprint",
                                               "source"],
                                        output="table"))

                    # d = dict(sshm.__keys__)

                    # print(_print_dict(d, format=_format))
                    msg = "info. OK."
                    Console.ok(msg)
                    return ""
                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem listing keys from ssh")

            elif arguments['--source'] in ['cm', 'cloudmesh', 'yaml']:

                try:
                    #sshm = SSHKeyManager()
                    d = Key.get_from_yaml(load_order=directory, store=False)

                    print(Printer.write(d,
                                        order=["name",
                                               "comment",
                                               "uri",
                                               "fingerprint",
                                               "source"],
                                        output=_format))
                    return ""
                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem listing keys from `{:}`".format(arguments['--source']))

            elif arguments['--source'] in ['git']:

                username = arguments["--username"]
                # print(username)
                if username == 'none':
                    conf = ConfigDict("cloudmesh.yaml")
                    username = conf["cloudmesh.github.username"]

                #sshm = SSHKeyManager()
                try:

                    d = Key.get_from_git(username, store=False)

                    print(Printer.write(d,
                                        order=["name",
                                               "comment",
                                               "uri",
                                               "fingerprint",
                                               "source"],
                                        output=_format))

                    msg = "info. OK."
                    Console.ok(msg)
                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem listing git keys from database")
                    return ""

            elif arguments['--source'] == 'db':

                try:
                    #sshdb = SSHKeyDBManager()
                    d = Key.all(output='dict')

                    if d is not None or d != []:

                        print(Printer.write(d,
                                            order=["name",
                                                   "comment",
                                                   "uri",
                                                   "fingerprint",
                                                   "source"],
                                            output=_format))

                        # print(_print_dict(d, output=arguments['--format']))
                        msg = "info. OK."
                        Console.ok(msg)
                    else:
                        Console.error("No keys in the database")
                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem listing keys from database")



        elif arguments['get']:

            try:
                name = arguments['NAME']
                #sshdb = SSHKeyDBManager()
                d = Key.all(output="dict")

                for key in d:
                    if key["name"] == name:
                        print("{:}: {:}".format(key['name'], key['fingerprint']))
                        msg = "info. OK."
                        Console.ok(msg)
                        return ""
                    else:
                        pass
                Console.error("The key is not in the database")
            except Exception as e:
                Error.traceback(e)
                Console.error("The key is not in the database")

        # key add [NAME] [--source=FILENAME]
        #     key add [NAME] [--git]

        elif arguments['add'] and arguments["--git"]:

            # Console.error("This feature is not yet implemented", traceflag=False)
            # return ""

            print('git add')
            #sshdb = SSHKeyDBManager()

            data = dotdict(arguments)

            keyname = data.NAME

            #
            # get name
            #
            conf = ConfigDict("cloudmesh.yaml")
            data.username = conf["cloudmesh.github.username"]
            data.name = arguments['NAME'] or data.username

            #
            # get git username
            #
            data.username = ConfigDict("cloudmesh.yaml")["cloudmesh.github.username"]

            if str(data.name).lower() in invalid_names:
                Console.error("The github user name is not set in the yaml file", traceflag=False)
                return ""

            try:
                Console.msg("Retrieving github ssh keys for user {username}".format(**data))
                #sshm = SSHKeyManager()
                Key.get_from_git(data.username)
                d = Key.all()
                # pprint(d)
            except Exception as e:
                Console.error("Problem adding keys to git for user: {username}".format(**data))
                return ""

            for key in d:
                if key is not None:
                    key["name"] = key["name"].replace("-", "_")
                    key["source"] = "git"
                    key["user"] = data.name
                    try:
                        o = dict(key)
                        o['value'] = key["string"]
                        Key.add_from_dict(key)
                    except Exception as e:
                        Console.error("The key {name} with that finger print already exists".format(**key),
                                      traceflag=False)

        elif arguments['add'] and not arguments["--git"]:

            #     key add [NAME] [--source=FILENAME]
            #     key add [NAME] [--git]

            #sshdb = SSHKeyDBManager()

            data = dotdict()

            #
            # get name
            #
            conf = ConfigDict("cloudmesh.yaml")
            data.username = conf["cloudmesh.profile.user"]
            data.name = arguments['NAME'] or data.username

            data.filename = arguments['--source']
            if data.filename == "db" or data.filename is None:
                data.filename = Config.path_expand("~/.ssh/id_rsa.pub")

            if str(data.name).lower() in invalid_names:
                msg = ("Your choice of keyname {name} is insufficient. \n"
                       "You must be chosing a keyname that is distingct on all clouds. \n"
                       "Possible choices are your gmail name, your XSEDE name, or \n"
                       "some name that is uniqe. "
                       "Best is also to set this name in \n"
                       "cloudmesh.profile.user as "
                       "part of your \n~/cloudmesh/cloudmesh.yaml file.")
                Console.error(msg.format(**data), traceflag=False)
                return ""

            try:
                Key.add_from_path(data.filename,
                                    data.name,
                                    source="ssh",
                                    uri="file://" + data.filename)
                print("Key {name} successfully added to the database".format(**data))
                msg = "info. OK."
                Console.ok(msg)

            except ValueError as e:
                Console.error("A key with this fingerprint already exists".format(**data), traceflag=False)
                Console.msg("Please use check with: key list")

            return ""

        elif arguments['default']:

            # print("default")

            if arguments['--select']:
                keyname = None
                try:
                    #sshdb = SSHKeyDBManager()
                    select = Key.select()
                    if select != 'q':
                        keyname = select.split(':')[0]
                        print("Setting key: {:} as default.".format(keyname))
                        Default.key = keyname
                        msg = "info. OK."
                        Console.ok(msg)
                except Exception as e:
                    Error.traceback(e)
                    Console.error("Setting default for selected key {:} failed.".format(keyname))

            else:
                try:
                    #sshdb = SSHKeyDBManager()
                    d = Key.table_dict()

                    for i in d:
                        if d[i]["is_default"] == "True":
                            key = d[i]
                            print("{:}: {:}".format(key['name'], key['fingerprint']))
                            msg = "info. OK."
                            Console.ok(msg)
                            return ""
                        else:
                            pass
                    Console.error("The key is not in the database")
                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem retrieving default key.")

        elif arguments['delete'] and arguments["--cloud"]:

            key = dotdict({
                'cloud': arguments["--cloud"],
                'name': arguments["NAME"]
            })
            try:
                Key.delete(key.name, key.cloud)
                msg = "info. OK."
                Console.ok(msg)
            except:
                Console.error("Problem deleting the key {name} on the cloud {cloud}".format(**key))

        elif arguments['delete']:

            # key delete (NAME | --select| --all)

            data = dotdict({
                'all': arguments['--all'] or False,
                'select': arguments['--select'] or False,
                'name': arguments['NAME'] or False,
            })


            pprint(data)
            # BUG delete all is not properly implemented

            if data.all:
                Console.TODO("Delete --all is not yet implemented.")
                # Key.delete()
            elif data.select:
                key = Key.select()
                print (key)

            else: # name
                Key.delete(data.name)


            msg = "info. OK."
            Console.ok(msg)

        elif arguments['upload']:

            # pprint(arguments)

            try:
                #
                # get username
                #
                conf = ConfigDict("cloudmesh.yaml")
                username = conf["cloudmesh"]["profile"]["user"]
                if username in ['None', 'TBD']:
                    username = None

                #
                # get cloudnames
                #
                clouds = []

                if arguments["--active"]:
                    cloud = 'active'
                else:
                    cloud = arguments["--cloud"] or Default.cloud

                if cloud == "all":
                    config = ConfigDict("cloudmesh.yaml")
                    clouds = config["cloudmesh"]["clouds"]
                elif cloud == 'active':
                    config = ConfigDict("cloudmesh.yaml")
                    clouds = config["cloudmesh"]["active"]
                else:
                    clouds.append(cloud)

                #
                # get keyname
                #

                for cloud in clouds:
                    status = 0
                    #sshdb = SSHKeyDBManager()
                    #sshm = SSHKeyManager()
                    keys = Key.all()
                    for key in keys:

                        print("upload key {} -> {}".format(key["name"],
                                                           cloud))

                        try:
                            status = Key.add_key_to_cloud(
                                username,
                                key["name"],
                                cloud)

                        except Exception as e:
                            Console.error("problem")
                            if "already exists" in str(e):
                                print("key already exists. Skipping upload. OK.")
                        if status == 1:
                            print("Problem uploading key {} to {}. failed.".format(key["name"],
                                                                                   cloud))

            except Exception as e:
                Console.error("Problem adding key to cloud")



from __future__ import print_function
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.default import Default
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.dotdict import dotdict
import os

class SetupCommand(PluginCommand, CloudPluginCommand):
    topics = {"setup": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command setup")

    # noinspection PyUnusedLocal
    @command
    def do_setup(self, args, arguments):
        """
        ::

            Usage:
                setup

            Examples:
                cm setup

        """
        arg = dotdict(arguments)


        #if Default.key in ["TBD", ""] or Default.user in ["TBD", ""]:
        os.system ("cm register profile")
        os.system ("cm reset")
        os.system ("cm key add --ssh")
        os.system ("cm info")

        return ""



from __future__ import print_function
from libcloud.compute.types import Provider
from libcloud.compute.providers import get_driver
import libcloud.security
import os
from pprint import pprint
from uuid import UUID
import re
from cloudmesh_client.common.todo import TODO
from cloudmesh_client.common.ConfigDict import Config, ConfigDict
from cloudmesh_client.common.FlatDict import FlatDict
from cloudmesh_client.cloud.iaas.CloudProviderBase import CloudProviderBase

from cloudmesh_client.cloud.iaas.provider.libcloud.CloudProviderLibcloud import CloudProviderLibcloud

from cloudmesh_client.shell.console import Console


class CloudProviderLibcloudEC2(CloudProviderLibcloud):
    def __init__(self, cloud_name, cloud_details, user=None, flat=True):
        super(CloudProviderLibcloudEC2, self).__init__(cloud_name, cloud_details, user=user)
        self.flat = flat
        self.cloud_type = "libcloud"
        self.kind = ["image", "vm", "flavor", "key"]
        self.cloudname = cloud_name
        self.initialize(cloud_name)

    def initialize(self, cloudname, user=None):

        Console.info("Initializing libcloud-ec2 for " + cloudname)
        cls = get_driver(Provider.EC2)

        d = ConfigDict("cloudmesh.yaml")
        self.config = d["cloudmesh"]["clouds"][cloudname]
        credentials = self.config["credentials"]
        cm_type = self.config["cm_type"]

        ec2_access_key = credentials['EC2_ACCESS_KEY']
        ec2_secret_key = credentials['EC2_SECRET_KEY']
        if not cloudname == "aws":
            auth_url = credentials["EC2_URL"]
            searchobj = re.match(r'^http[s]?://(.+):([0-9]+)/([a-zA-Z/]*)',
                                 auth_url,
                                 re.M | re.I)
            path = None
            host = None
            port = None
            if searchobj:
                host = searchobj.group(1)
                port = searchobj.group(2)
                path = searchobj.group(3)

                Console.info("url : " + searchobj.group())
                Console.info("host: " + host)
                Console.info("port: " + port)
                Console.info("path: " + path)

                extra_args = {'path': path}
            else:
                Console.error("Authentication url incorrect: {}".format(auth_url))

            self.provider = cls(ec2_access_key,
                                ec2_secret_key,
                                host=host,
                                port=port,
                                **extra_args)
        else:
            Console.info("AWS INIT")
            self.provider = cls(ec2_access_key, ec2_secret_key)

    def create_node(self, **kwargs):
        """
        Create a VM instance.

        Reference: https://libcloud.readthedocs.io/en/latest/_modules/libcloud/compute/drivers/ec2.html#BaseEC2NodeDriver.create_node

        :keyword    name: the name of VM
        :type       name: ``str``

        :keyword    image: the name of vm image
        :type       name: :class:`NodeImage`

        :keyword    size: the size of vm instance
        :type       size: :class:`NodeSize`

        :keyword    ex_keyname: the keyname
        :type       name: ``str``
        """

        pprint("create_node call in aws")
        self.provider.create_node(name=kwargs['name'], image=kwargs['image'], size=kwargs['size'])

    def get_ips(self, name):

        vlist = self.provider.list_nodes()

        for vm in vlist:
            # [<Node: uuid=3793375fbc6577391486b0ae43b1d1fb303aa0c1, name=TBD-001, state=RUNNING, public_ips=['54.205.149.43'], private_ips=['10.164.162.29'], provider=Amazon EC2 ...>]
            if name == vm.uuid:
                return vm.public_ips[0]
        for vm in vlist:
            if name == vm.name:
                return vm.public_ips[0]
        return NULL
        



from __future__ import print_function
from cloudmesh_client.default import Default
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.cloud.limits import Limits
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.ConfigDict import ConfigDict
import os

class LoadCommand(PluginCommand, CloudPluginCommand):
    topics = {"load": "shell"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command load")

    # noinspection PyUnusedLocal
    @command
    def do_load(self, args, arguments):
        """
        ::

            Usage:
                load MODULE [PYPI]


            ARGUMENTS:
               MODULE  The name of the module

            PREDEFINED MODULE NAMES:
               vbox    loads vbox command

            Examples:
                cm load cloudmesh_vagrant.cm_vbox.do_vbox
                    lists the plugins currently loaded

        """
        arg = dotdict(arguments)

        # importlib.import_module('matplotlib.text')

        plugins = ConfigDict(filename="cloudmesh.yaml")

        if arg.MODULE == "vbox":
            arg.MODULE = "cloudmesh_vagrant.cm_vbox.do_vbox"
            arg.PYPI = "cloudmesh_vagrant"

        if arg.PYPI is not None:
            try:
                import cloudmesh_vagrant
            except:
                os.system("pip install cloudmesh_vagrant")


        try:
                print("LOADING ->", arg.MODULE)
                self.load_instancemethod(arg.MODULE)

        except:
            Console.error("Problem loading module {}".format(arg.MODULE),
                          traceflag=True)
        return ""






""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_azure/test_azure_cli.py:Test_flavor.test_001

nosetests -v --nocapture tests/cm_azure/test_azure_cli.py

or

nosetests -v tests/cm_azure/test_azure_cli.py

"""

import os
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from pprint import pprint

class Test_azure_cli:

    data = dotdict({
        "cloud": "azure",
    })

    def run(self, command):
        self.cloud = "azure"
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return result

    def setup(self):
        HEADING()
        result = self.run("cm default cloud=azure")
        assert "ok." in result

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_001(self):
        """
            Test VM refresh
        :return:
        """
        HEADING()
        result = self.run("cm vm refresh")
        assert "Refresh VMs for cloud azure" in result

    def test_002(self):
        """
            Test VM list
        :return:
        """
        HEADING()
        result = self.run("cm vm list")
        assert "Listing VMs on Cloud: azure" in result

    def test_003(self):
        """
            Test Image refresh
        :return:
        """
        HEADING()
        result = self.run("cm image refresh")
        assert "Refresh image for cloud azure" in result

    def test_004(self):
        """
            Test Image list
        :return:
        """
        HEADING()
        result = self.run("cm image list")
        assert "image_family" in result

    def test_005(self):
        """
            Test Flavor refresh
        :return:
        """
        HEADING()
        result = self.run("cm flavor refresh")
        assert "Refresh flavor for cloud azure" in result

    def test_006(self):
        """
            Test Flavor list
        :return:
        """
        HEADING()
        result = self.run("cm flavor list")
        assert "web_worker_resource_disk_size" in result

    def test_007(self):
        """
            Test AKEY command test
        :return:
        """
        HEADING()
        # os.getcwd()
        pprint("This test will work only if mycer.pub,mycer.pem and mycer.pfx is copied to home directory")
        pprint("Current Working directory::"+os.getcwd())
        key_name = "test-key"
        pub_key_path = "~/mycer.pub"
        certificate_path = "~/mycer.pem"
        pfx_file_path = "~/mycer.pfx"
        cm_akey_command = "cm akey add --name="+key_name+" --pub="+pub_key_path+" --cert="+certificate_path+" --pfx="+pfx_file_path
        result = self.run(cm_akey_command)
        assert "Azure key added" in result

    def test_008(self):
        """
            Test VM create command
        :return:
        """
        HEADING()
        result = self.run("cm vm boot")
        assert "VM boot up successful" in result


    def test_009(self):
        """
            Test VM delete command
        :return:
        """
        HEADING()
        result = self.run("cm vm delete")
        assert "Delete Success" in result



from cloudmesh_client.cloud.hpc.provider.slurm.BatchProviderSLURM import BatchProviderSLURM
from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.Error import Error


# noinspection PyPep8Naming
def BatchProvider(name, user=None):
    """
    Returns a provider for a given batch system that is defined
    in the cloudmesh.yaml file. Here you find a section such as::

        hpc:
            active:
            - india
            clusters:
              india:
                cm_heading: India HPC CLuster
                cm_host: india
                cm_label: indiahpc
                cm_type: slurm
                cm_type_version: 14.11.9
                credentials:
                  username: TBD
                  project: None
                default:
                  queue: delta
                  experiment_dir: /N/u/{username}/experiment
                  prefix:
                    username: null

    You can define a new resource by copying this section under india and
    give it a name according to your cluster. Fill out the username,
    and if appropirate define a project, which is the
    account this queue is charged under

    Ussage::

      provider = BatchProvider("india")

    Additional functions are defined as it inherits from BatchProviderBase
    and the provider type which is in the above example slurm

    :param name: name of the cluster in the cloudmesh yaml file
    :param user: cloudmesh user name associated with this queue.
    :return: the provider
    """
    try:
        d = ConfigDict("cloudmesh.yaml")
        details = d["cloudmesh"]["hpc"]["clusters"][name]

        if details["cm_type"].lower() in ["slurm"]:
            return BatchProviderSLURM()
        else:
            ValueError("batch  provider not supported.")

    except Exception as e:
        Error.traceback(e)



from __future__ import print_function
from cloudmesh_client.cloud.flavor import Flavor
from cloudmesh_client.shell.command import command, PluginCommand, CloudPluginCommand
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default


class FlavorCommand(PluginCommand, CloudPluginCommand):
    topics = {"flavor": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command flavor")

    # noinspection PyUnusedLocal
    @command
    def do_flavor(self, args, arguments):
        """
        ::

            Usage:
                flavor refresh [--cloud=CLOUD] [-v]
                flavor list [ID] [--cloud=CLOUD] [--format=FORMAT] [--refresh] [-v]

                This lists out the flavors present for a cloud

            Options:
               --format=FORMAT  the output format [default: table]
               --cloud=CLOUD    the cloud name
               --refresh        refreshes the data before displaying it
                                from the cloud

            Examples:
                cm flavor refresh
                cm flavor list
                cm flavor list --format=csv
                cm flavor show 58c9552c-8d93-42c0-9dea-5f48d90a3188 --refresh

        """

        cloud = arguments["--cloud"] or Default.cloud
        if cloud is None:
            Console.error("Default cloud doesn't exist")
            return

        if arguments["-v"]:
            print("Cloud: {}".format(cloud))

        if arguments["refresh"] or Default.refresh:
            msg = "Refresh flavor for cloud {:}.".format(cloud)
            if Flavor.refresh(cloud):
                Console.ok("{:} ok".format(msg))
            else:
                Console.error("{:} failed".format(msg))
                return ""

        if arguments["list"]:

            id = arguments['ID']
            live = arguments['--refresh']
            output_format = arguments["--format"]

            counter = 0

            result = None
            while counter < 2:
                if id is None:
                    result = Flavor.list(cloud, output_format)
                else:
                    result = Flavor.details(cloud, id, live, output_format)
                if counter == 0 and result is None:
                    if not Flavor.refresh(cloud):
                        msg = "Refresh flavor for cloud {:}.".format(cloud)
                        Console.error("{:} failed.".format(msg))
                counter += 1

            if result is None:
                Console.error("No flavor(s) found. Failed.")
            else:
                print(result)
            return ""
